# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import unicode_literals

import os
import tempfile
import shutil
import json
import unittest
from decimal import Decimal

import numpy as np
from scipy.io import savemat, loadmat
from scipy.sparse.csc import csc_matrix

from six import BytesIO

from psamm import importer
from psamm.importer import ModelLoadError
from psamm.datasource.native import NativeModel
from psamm.datasource.entry import (DictCompoundEntry as CompoundEntry,
                                    DictReactionEntry as ReactionEntry,
                                    DictCompartmentEntry as CompartmentEntry)
from psamm.datasource.reaction import parse_reaction
from psamm.reaction import Reaction, Compound
from psamm.formula import Formula
from psamm.expression import boolean

from psamm.importers.sbml import StrictImporter as SBMLStrictImporter
from psamm.importers.sbml import NonstrictImporter as SBMLNonstrictImporter
from psamm.importers.cobrajson import Importer as CobraJSONImporter
from psamm.importers.matlab import Importer as MatlabImporter


class TestImporterBaseClass(unittest.TestCase):
    def setUp(self):
        self.importer = importer.Importer()

    def test_try_parse_formula(self):
        result = self.importer._try_parse_formula('cpd_1', ' ')
        self.assertIsNone(result)

        # Should not return parsed result!
        result = self.importer._try_parse_formula('cpd_1', 'CO2')
        self.assertEqual(result, 'CO2')

        result = self.importer._try_parse_formula('cpd_1', 'XYZ...\u03c0')
        self.assertEqual(result, 'XYZ...\u03c0')

    def test_try_parse_reaction(self):
        result = self.importer._try_parse_reaction('rxn_1', 'A => (3) B')
        self.assertIsInstance(result, Reaction)

        with self.assertRaises(importer.ParseError):
            self.importer._try_parse_reaction('rxn_1', '+ + ==> \u03c0 5')

    def test_try_parse_gene_association(self):
        result = self.importer._try_parse_gene_association(
            'rxn_1', 'g1 and (g2 or g3)')
        self.assertIsInstance(result, boolean.Expression)

        result = self.importer._try_parse_gene_association('rxn_1', ' ')
        self.assertIsNone(result)

        result = self.importer._try_parse_gene_association('rxn_1', '123 and')
        self.assertEqual(result, '123 and')


class TestImporterCheckAndWriteModel(unittest.TestCase):
    def setUp(self):
        self.dest = tempfile.mkdtemp()

        self.model = NativeModel({
            'id': 'test_mode',
            'name': 'Test model',
        })

        # Compounds
        self.model.compounds.add_entry(CompoundEntry({
            'id': 'cpd_1',
            'formula': Formula.parse('CO2')
        }))
        self.model.compounds.add_entry(CompoundEntry({
            'id': 'cpd_2',
        }))
        self.model.compounds.add_entry(CompoundEntry({
            'id': 'cpd_3',
        }))

        # Compartments
        self.model.compartments.add_entry(CompartmentEntry({
            'id': 'c'
        }))
        self.model.compartments.add_entry(CompartmentEntry({
            'id': 'e'
        }))

        # Reactions
        self.model.reactions.add_entry(ReactionEntry({
            'id': 'rxn_1',
            'equation': parse_reaction('(2) cpd_1[c] <=> cpd_2[e] + cpd_3[c]'),
            'genes': boolean.Expression('g_1 and (g_2 or g_3)'),
            'subsystem': 'Some subsystem'
        }))
        self.model.reactions.add_entry(ReactionEntry({
            'id': 'rxn_2',
            'equation': parse_reaction(
                '(0.234) cpd_1[c] + (1.34) cpd_2[c] => cpd_3[c]'),
            'subsystem': 'Biomass'
        }))

        self.model.biomass_reaction = 'rxn_2'
        self.model.limits['rxn_1'] = 'rxn_1', Decimal(-42), Decimal(1000)
        self.model.exchange[Compound('cpd_2', 'e')] = (
            Compound('cpd_2', 'e'), 'EX_cpd_2', -10, 0)

    def tearDown(self):
        shutil.rmtree(self.dest)

    def test_get_default_compartment(self):
        importer.get_default_compartment(self.model)

    def test_detect_best_flux_limit(self):
        flux_limit = importer.detect_best_flux_limit(self.model)
        self.assertEqual(flux_limit, 1000)

    def test_infer_compartment_entries(self):
        self.model.compartments.clear()
        importer.infer_compartment_entries(self.model)

        self.assertEqual(len(self.model.compartments), 2)
        self.assertEqual({c.id for c in self.model.compartments}, {'c', 'e'})

    def test_get_output_limit(self):
        self.assertIsNone(importer._get_output_limit(1000, 1000))
        self.assertEqual(importer._get_output_limit(-5, 0), -5)

    def test_generate_limit_items(self):
        item = dict(importer._generate_limit_items(None, None))
        self.assertEqual(item, {})

        item = dict(importer._generate_limit_items(None, 400))
        self.assertEqual(item, {'upper': 400})

        item = dict(importer._generate_limit_items(-56, None))
        self.assertEqual(item, {'lower': -56})

        item = dict(importer._generate_limit_items(-224, 0))
        self.assertEqual(item, {'lower': -224, 'upper': 0})

        item = dict(importer._generate_limit_items(-5, -5))
        self.assertEqual(item, {'fixed': -5})

    def test_count_genes(self):
        self.assertEqual(importer.count_genes(self.model), 3)

    def test_write_yaml_model(self):
        importer.write_yaml_model(self.model, self.dest)


class TestSBMLImporter(unittest.TestCase):
    def setUp(self):
        self.dest = tempfile.mkdtemp()
        with open(os.path.join(self.dest, 'model.sbml'), 'wb') as f:
            f.write('''<?xml version="1.0" encoding="UTF-8"?>
    <sbml xmlns="http://www.sbml.org/sbml/level3/version1/core"
          xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2"
          xmlns:html="http://www.w3.org/1999/xhtml"
          level="3" version="1"
          fbc:required="false">
     <model id="test_model" name="Test model">
      <notes>
       <body xmlns="http://www.w3.org/1999/xhtml">
        <p>This is model information intended to be seen by humans.</p>
       </body>
      </notes>
      <listOfCompartments>
       <compartment id="C_c" name="cell" constant="true"/>
       <compartment id="C_b" name="boundary" constant="true"/>
      </listOfCompartments>
      <listOfSpecies>
       <species id="M_Glucose_LPAREN_c_RPAREN_"
                metaid="meta_M_Glucose_LPAREN_c_RPAREN_" name="Glucose"
                compartment="C_c" constant="false" boundaryCondition="false"
                hasOnlySubstanceUnits="false" fbc:charge="0"
                fbc:chemicalFormula="C6H12O6">
        <notes>
         <body xmlns="http://www.w3.org/1999/xhtml">
          <p>This is compound information intended to be seen by humans.</p>
         </body>
        </notes>
       </species>
       <species id="M_Glucose_6_DASH_P_LPAREN_c_RPAREN_" name="Glucose-6-P"
                compartment="C_c" constant="false" boundaryCondition="false"
                hasOnlySubstanceUnits="false" fbc:charge="-2"
                fbc:chemicalFormula="C6H11O9P"/>
       <species id="M_H2O_LPAREN_c_RPAREN_" name="H2O" compartment="C_c"
                constant="false" boundaryCondition="false"
                hasOnlySubstanceUnits="false" fbc:charge="0"
                fbc:chemicalFormula="H2O"/>
       <species id="M_Phosphate_LPAREN_c_RPAREN_" name="Phosphate"
                compartment="C_c" constant="false" boundaryCondition="false"
                hasOnlySubstanceUnits="false" fbc:charge="-2"
                fbc:chemicalFormula="HO4P"/>
       <species id="M_Biomass" name="Biomass" compartment="C_b" constant="false"
                boundaryCondition="true" hasOnlySubstanceUnits="false"/>
      </listOfSpecies>
      <listOfParameters>
       <parameter constant="true" id="P_lower_G6Pase" value="-10"/>
       <parameter constant="true" id="P_lower_zero" value="0"/>
       <parameter constant="true" id="P_upper_default" value="1000"/>
      </listOfParameters>
      <listOfReactions>
       <reaction id="R_G6Pase" metaid="meta_R_G6Pase" reversible="true"
                 fast="false" fbc:lowerFluxBound="P_lower_G6Pase"
                 fbc:upperFluxBound="P_upper_default">
        <listOfReactants>
         <speciesReference species="M_Glucose_LPAREN_c_RPAREN_" stoichiometry="2"
                           constant="true"/>
         <speciesReference species="M_Phosphate_LPAREN_c_RPAREN_" stoichiometry="2"
                           constant="true"/>
        </listOfReactants>
        <listOfProducts>
         <speciesReference species="M_H2O_LPAREN_c_RPAREN_" stoichiometry="2"
                           constant="true"/>
         <speciesReference species="M_Glucose_6_DASH_P_LPAREN_c_RPAREN_"
                           stoichiometry="2" constant="true"/>
        </listOfProducts>
        <notes>
         <body xmlns="http://www.w3.org/1999/xhtml">
          <p>This is reaction information intended to be seen by humans.</p>
         </body>
        </notes>
        <annotation>
         <RDF xmlns="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
              xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
          <Description about="#meta_R_G6Pase">
           <bqbiol:isVersionOf>
            <Bag>
             <li resource="http://identifiers.org/ec-code/3.1.3.9"/>
            </Bag>
           </bqbiol:isVersionOf>
          </Description>
         </RDF>
        </annotation>
       </reaction>
       <reaction id="R_Biomass" reversible="false" fast="false"
                 fbc:lowerFluxBound="P_lower_zero"
                 fbc:upperFluxBound="P_upper_default">
        <listOfReactants>
         <speciesReference species="M_Glucose_6_DASH_P_LPAREN_c_RPAREN_"
                           stoichiometry="0.56" constant="true"/>
         <speciesReference species="M_Glucose_LPAREN_c_RPAREN_"
                           stoichiometry="0.88" constant="true"/>
        </listOfReactants>
        <listOfProducts>
         <speciesReference species="M_Biomass" stoichiometry="1" constant="true"/>
        </listOfProducts>
       </reaction>
      </listOfReactions>
      <fbc:listOfObjectives fbc:activeObjective="obj1">
       <fbc:objective fbc:id="obj1" fbc:name="Objective 1" fbc:type="maximize">
        <fbc:listOfFluxObjectives>
         <fbc:fluxObjective fbc:reaction="R_Biomass" fbc:coefficient="1"/>
        </fbc:listOfFluxObjectives>
       </fbc:objective>
      </fbc:listOfObjectives>
     </model>
    </sbml>'''.encode('utf-8'))

    def tearDown(self):
        shutil.rmtree(self.dest)

    def test_create_strict_importer(self):
        importer = SBMLStrictImporter()
        importer.help()
        importer.import_model(os.path.join(self.dest, 'model.sbml'))

    def test_strict_importer_main(self):
        output_dir = os.path.join(self.dest, 'output')
        os.mkdir(output_dir)
        importer.main(importer_class=SBMLStrictImporter, args=[
            '--source', os.path.join(self.dest, 'model.sbml'),
            '--dest', output_dir
        ])

    def test_create_nonstrict_importer(self):
        importer = SBMLNonstrictImporter()
        importer.help()
        importer.import_model(os.path.join(self.dest, 'model.sbml'))

    def test_nonstrict_importer_main(self):
        output_dir = os.path.join(self.dest, 'output')
        os.mkdir(output_dir)
        importer.main(importer_class=SBMLNonstrictImporter, args=[
            '--source', os.path.join(self.dest, 'model.sbml'),
            '--dest', output_dir
        ])


class TestCobraJSONImporter(unittest.TestCase):
    def setUp(self):
        self.dest = tempfile.mkdtemp()
        with open(os.path.join(self.dest, 'model.json'), 'w') as f:
            json.dump({
                'id': 'test_model',
                'compartments': {
                    'c': 'Cytosol',
                    'e': 'Extracellular',
                },
                'metabolites': [
                    {
                        'id': 'cpd_1',
                        'name': 'Compound 1',
                        'formula': 'CO2',
                        'compartment': 'c'
                    },
                    {
                        'id': 'cpd_2',
                        'name': 'Compound 2',
                        'charge': -2,
                        'compartment': 'e'
                    },
                    {
                        'id': 'cpd_3',
                        'compartment': 'c'
                    }
                ],
                'reactions': [
                    {
                        'id': 'rxn_1',
                        'metabolites': {
                            'cpd_1': -1,
                            'cpd_2': 2,
                            'cpd_3': 1,
                        },
                        'lower_bound': -20,
                        'upper_bound': 100,
                        'gene_reaction_rule': 'gene_1 and (gene_2 or gene_3)'
                    }
                ]
            }, f)

    def tearDown(self):
        shutil.rmtree(self.dest)

    def test_create_importer(self):
        importer = CobraJSONImporter()
        importer.help()
        importer.import_model(os.path.join(self.dest, 'model.json'))

    def test_cobra_json_main(self):
        output_dir = os.path.join(self.dest, 'output')
        os.mkdir(output_dir)
        importer.main(importer_class=CobraJSONImporter, args=[
            '--source', os.path.join(self.dest, 'model.json'),
            '--dest', output_dir
        ])

    def test_cobra_json_main_bigg(self):
        def mock_urlopen(url):
            return open(os.path.join(self.dest, 'model.json'), 'rb')

        output_dir = os.path.join(self.dest, 'output')
        os.mkdir(output_dir)
        importer.main_bigg(
            args=['mock_model', '--dest', output_dir], urlopen=mock_urlopen)


class TestBiGGImportMain(unittest.TestCase):
    def test_list_models(self):
        def mock_urlopen(url):
            return BytesIO(b'''
{"results_count": 4,
 "results": [
   {"organism": "Escherichia coli str. K-12 substr. MG1655",
    "metabolite_count": 72,
    "bigg_id": "e_coli_core",
    "gene_count": 137,
    "reaction_count": 95},
   {"organism": "Homo sapiens",
    "metabolite_count": 342,
    "bigg_id": "iAB_RBC_283",
    "gene_count": 346,
    "reaction_count": 469},
   {"organism": "Escherichia coli str. K-12 substr. MG1655",
    "metabolite_count": 1668,
    "bigg_id": "iAF1260",
    "gene_count": 1261,
    "reaction_count": 2382}]}''')

        importer.main_bigg(['list'], urlopen=mock_urlopen)


class TestMatlabImporter(unittest.TestCase):
    def setUp(self):
        self.dest = tempfile.mkdtemp()
        dt = np.dtype([
            ('comps', 'O'),
            ('compNames', 'O'),
            ('rxns', 'O'),
            ('rxnNames', 'O'),
            ('rxnGeneMat', 'O'),
            ('ub', 'O'),
            ('lb', 'O'),
            ('S', 'O'),
            ('grRules', 'O'),
            ('subSystems', 'O'),
            ('mets', 'O'),
            ('metFormulas', 'O'),
            ('metNames', 'O'),
            ('metCharges', 'O'),
            ('metNotes', 'O'),
            ('c', 'O'),
        ])
        model1 = np.ndarray((1, 1), dtype=dt)
        # compartments
        model1['comps'][0, 0] = np.ndarray((3, 1), dtype='O')
        model1['comps'][0, 0][0][0] = np.array(['c'])
        model1['comps'][0, 0][1][0] = np.array(['e'])
        model1['comps'][0, 0][2][0] = np.array(['b'])
        model1['compNames'][0, 0] = np.ndarray((3, 1), dtype='O')
        model1['compNames'][0, 0][0][0] = np.array(['Cell'])
        model1['compNames'][0, 0][1][0] = np.array(['Extracellular'])
        model1['compNames'][0, 0][2][0] = np.array(['Boundary'])
        # reactions
        model1['rxns'][0, 0] = np.ndarray((3, 1), dtype='O')
        model1['rxns'][0, 0][0][0] = np.array(['rxn1'])
        model1['rxns'][0, 0][1][0] = np.array(['rxn2'])
        model1['rxns'][0, 0][2][0] = np.array(['rxn3'])
        model1['rxnNames'][0, 0] = np.ndarray((3, 1), dtype='O')
        model1['rxnNames'][0, 0][0][0] = np.array(['rxn1'])
        model1['rxnNames'][0, 0][1][0] = np.array([])
        model1['rxnNames'][0, 0][2][0] = np.array(['rxn3'])
        model1['rxnGeneMat'][0, 0] = np.ndarray((3, 1), dtype='O')
        model1['rxnGeneMat'][0, 0][0][0] = np.array([])
        model1['rxnGeneMat'][0, 0][1][0] = np.array([0])
        model1['rxnGeneMat'][0, 0][2][0] = np.array([0])
        model1['ub'][0, 0] = np.ndarray((3, 1), dtype='O')
        model1['ub'][0, 0][0][0] = np.array([1000])
        model1['ub'][0, 0][1][0] = np.array([0])
        model1['ub'][0, 0][2][0] = np.array([1000])
        model1['lb'][0, 0] = np.ndarray((3, 1), dtype='O')
        model1['lb'][0, 0][0][0] = np.array([0])
        model1['lb'][0, 0][1][0] = np.array([-1000])
        model1['lb'][0, 0][2][0] = np.array([-1000])
        # equation matrix
        data = np.array([-1, -2, 1, -1, 2, -1, 1])
        row = np.array([0, 1, 2, 2, 3, 3, 4])
        col = np.array([0, 0, 0, 1, 1, 2, 2])
        model1['S'][0, 0] = csc_matrix((data, (row, col)), shape=(5, 3))
        # genes
        model1['grRules'][0, 0] = np.ndarray((3, 1), dtype='O')
        model1['grRules'][0, 0][0][0] = np.array(['g1 and g2'])
        model1['grRules'][0, 0][1][0] = np.array(['g3 or g4'])
        model1['grRules'][0, 0][2][0] = np.array([])
        # subsystems
        model1['subSystems'][0, 0] = np.ndarray((3, 1), dtype='O')
        model1['subSystems'][0, 0][0][0] = np.array(['system1'])
        model1['subSystems'][0, 0][1][0] = np.ndarray((1, 1), 'O')
        model1['subSystems'][0, 0][1][0][0][0] = np.array(['system2'])
        model1['subSystems'][0, 0][2][0] = np.array(['system3'])
        # compounds
        model1['mets'][0, 0] = np.ndarray((5, 1), dtype='O')
        model1['mets'][0, 0][0][0] = np.array(['A[e]'])
        model1['mets'][0, 0][1][0] = np.array(['B_c'])
        model1['mets'][0, 0][2][0] = np.array(['C(c)'])
        model1['mets'][0, 0][3][0] = np.array(['D[c]'])
        model1['mets'][0, 0][4][0] = np.array(['E001e'])
        model1['metFormulas'][0, 0] = np.ndarray((5, 1), dtype='O')
        model1['metFormulas'][0, 0][0][0] = np.array(['H2O'])
        model1['metFormulas'][0, 0][1][0] = np.array(['C5H10O2'])
        model1['metFormulas'][0, 0][2][0] = np.array(['H2SPO4'])
        model1['metFormulas'][0, 0][3][0] = np.array(['CH4N2'])
        model1['metFormulas'][0, 0][4][0] = np.array(['CO2'])
        model1['metNames'][0, 0] = np.ndarray((5, 1), dtype='O')
        model1['metNames'][0, 0][0][0] = np.array(['A'])
        model1['metNames'][0, 0][1][0] = np.array(['B'])
        model1['metNames'][0, 0][2][0] = np.array(['C'])
        model1['metNames'][0, 0][3][0] = np.array(['D'])
        model1['metNames'][0, 0][4][0] = np.array(['E'])
        model1['metCharges'][0, 0] = np.ndarray((5, 1), dtype='O')
        model1['metCharges'][0, 0][0][0] = 1
        model1['metCharges'][0, 0][1][0] = 2
        model1['metCharges'][0, 0][2][0] = -1.0
        model1['metCharges'][0, 0][3][0] = 0
        model1['metCharges'][0, 0][4][0] = 0
        model1['metNotes'][0, 0] = np.ndarray((5, 1), dtype='O')
        model1['metNotes'][0, 0][0][0] = np.array(['A'])
        model1['metNotes'][0, 0][1][0] = np.array(['B'])
        model1['metNotes'][0, 0][2][0] = np.array([])
        model1['metNotes'][0, 0][3][0] = np.array(['D'])
        model1['metNotes'][0, 0][4][0] = np.array(['E'])
        # biomass reaction
        model1['c'][0, 0] = np.ndarray((3, 1), dtype='O')
        model1['c'][0, 0][0][0] = 0
        model1['c'][0, 0][1][0] = 0
        model1['c'][0, 0][2][0] = 1

        os.mkdir(os.path.join(self.dest, 'models'))
        savemat(os.path.join(self.dest, 'model1.mat'), {'model1': model1})
        savemat(os.path.join(self.dest, 'models', 'model1.mat'),
                {'model1': model1})
        model2 = model1.copy()
        model2['c'][0, 0][1][0] = 1
        savemat(os.path.join(self.dest, 'models', 'model2.mat'),
                {'model1': model1, 'model2': model2})

    def tearDown(self):
        shutil.rmtree(self.dest)

    def test_create_importer(self):
        importer = MatlabImporter()
        importer.help()
        importer.import_model(os.path.join(self.dest, 'model1.mat'))
        importer.import_model(self.dest)
        importer.import_model(os.path.join(self.dest, 'models', 'model2.mat'),
                              'model2')
        with self.assertRaises(ModelLoadError):
            importer.import_model(os.path.join(self.dest, 'models'))
        with self.assertRaises(ModelLoadError):
            importer.import_model(os.path.join(
                self.dest, 'models', 'model2.mat'))

    def test_matlab_importer_main(self):
        output_dir = os.path.join(self.dest, 'output')
        os.mkdir(output_dir)
        importer.main(importer_class=MatlabImporter,
                      args=[
                          '--source', os.path.join(self.dest, 'model1.mat'),
                          '--dest', output_dir,
                      ])



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Parser and representation of chemical formulas.

Chemical formulas (:class:`.Formula`) are represented as a number of
:class:`FormulaElements <.FormulaElement>` with associated counts. A
:class:`.Formula` is itself a :class:`.FormulaElement` so a formula can contain
subformulas. This allows some simple structure to be represented.
"""

from __future__ import unicode_literals

import re
from collections import Counter
import functools
import operator
import numbers

import six
from six import iteritems
from six.moves import reduce

from .expression.affine import Expression


class FormulaElement(object):
    """Base class representing elements of a formula"""

    def __add__(self, other):
        """Add formula elements creating subformulas"""
        if isinstance(other, FormulaElement):
            if self == other:
                return Formula({self: 2})
            return Formula({self: 1, other: 1})
        return NotImplemented

    def __radd__(self, other):
        return self + other

    def __or__(self, other):
        """Merge formula elements into one formula"""
        return Formula({self: 1}) | other

    def __ror__(self, other):
        return self | other

    def __mul__(self, other):
        """Multiply formula element by other"""
        return Formula({self: other})

    def __rmul__(self, other):
        return self * other

    def repeat(self, count):
        """Repeat formula element by creating a subformula"""
        return Formula({self: count})

    def variables(self):
        """Iterator over variables in formula element"""
        return iter([])

    def substitute(self, mapping):
        """Return formula element with substitutions performed"""
        return self


class _AtomType(type):
    """Metaclass that gives the Atom class properties for each element.

    A class based on this metaclass (i.e. :class:`.Atom`) will have singleton
    elements with each name, and will have a property for each element which
    contains the instance for that element.
    """

    _ELEMENTS = set([
        'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al',
        'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn',
        'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb',
        'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In',
        'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm',
        'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta',
        'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At',
        'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk',
        'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt',
        'Ds', 'Rg', 'Cn', 'Uut', 'Fl', 'Uup', 'Lv', 'Uus', 'Uuo'
    ])

    _instances = {}

    def __call__(self, name, *args, **kwargs):
        instances = _AtomType._instances.setdefault(self, {})
        if name not in instances:
            instances[name] = super(_AtomType, self).__call__(
                name, *args, **kwargs)
        return instances[name]

    def __getattribute__(self, name):
        if name in _AtomType._ELEMENTS:
            return self(name)
        return super(_AtomType, self).__getattribute__(name)


@six.python_2_unicode_compatible
@functools.total_ordering
@six.add_metaclass(_AtomType)
class Atom(FormulaElement):
    """Represent an atom in a chemical formula

    >>> hydrogen = Atom.H
    >>> oxygen = Atom.O
    >>> str(oxygen | 2*hydrogen)
    'H2O'
    """

    def __init__(self, symbol):
        self._symbol = symbol

    @property
    def symbol(self):
        """Atom symbol

        >>> Atom.H.symbol
        'H'
        """

        return self._symbol

    def __hash__(self):
        return hash('Atom') ^ hash(self._symbol)

    def __eq__(self, other):
        return isinstance(other, Atom) and self._symbol == other._symbol

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        if isinstance(other, Atom):
            return self._symbol < other._symbol
        return NotImplemented

    def __str__(self):
        return self._symbol

    def __repr__(self):
        return str('Atom({})').format(repr(self._symbol))


@six.python_2_unicode_compatible
class Radical(FormulaElement):
    """Represents a radical or other unknown subformula"""

    def __init__(self, symbol):
        self._symbol = symbol

    @property
    def symbol(self):
        """Radical symbol

        >>> Radical('R1').symbol
        'R1'
        """

        return self._symbol

    def __hash__(self):
        return hash('Radical') ^ hash(self._symbol)

    def __eq__(self, other):
        return isinstance(other, Radical) and self._symbol == other._symbol

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return self._symbol

    def __repr__(self):
        return str('Radical({})').format(repr(self._symbol))


@six.python_2_unicode_compatible
class Formula(FormulaElement):
    """Representation of a chemial formula

    This is represented as a number of
    :class:`FormulaElements <.FormulaElement>` with associated counts.

    >>> f = Formula({Atom.C: 6, Atom.H: 12, Atom.O: 6})
    >>> str(f)
    'C6H12O6'
    """

    def __init__(self, values={}):
        self._values = Counter()
        self._variables = set()

        for element, value in iteritems(values):
            if not isinstance(element, FormulaElement):
                raise ValueError('Not a formula element: {}'.format(
                    repr(element)))
            if (value != 0 and
                    (not isinstance(element, Formula) or len(element) > 0)):
                self._values[element] = value

            if callable(getattr(value, 'variables', None)):
                for var in value.variables():
                    self._variables.add(var)
            for var in element.variables():
                self._variables.add(var)

    def substitute(self, mapping):
        result = self.__class__()
        for element, value in iteritems(self._values):
            if callable(getattr(value, 'substitute', None)):
                value = value.substitute(mapping)
                if isinstance(value, int) and value <= 0:
                    raise ValueError(
                        'Expression evaluated to non-positive number')
            # TODO does not merge correctly with subformulas
            result += value * element.substitute(mapping)
        return result

    def flattened(self):
        """Return formula where subformulas have been flattened

        >>> str(Formula.parse('(CH2)(CH2)2').flattened())
        'C3H6'
        """

        stack = [(self, 1)]
        result = Counter()
        while len(stack) > 0:
            var, value = stack.pop()
            if isinstance(var, Formula):
                for sub_var, sub_value in iteritems(var._values):
                    stack.append((sub_var, value*sub_value))
            else:
                result[var] += value
        return Formula(result)

    def variables(self):
        return iter(self._variables)

    def __iter__(self):
        return iter(self._values)

    def items(self):
        """Iterate over (:class:`.FormulaElement`, value)-pairs"""
        return iteritems(self._values)

    def is_variable(self):
        return len(self._variables) > 0

    def __contains__(self, element):
        return element in self._values

    def get(self, element, default=None):
        """Return value for element or default if not in the formula."""
        return self._values.get(element, default)

    def __getitem__(self, element):
        if element not in self._values:
            raise KeyError(repr(element))
        return self._values[element]

    def __len__(self):
        return len(self._values)

    def __str__(self):
        """Return formula represented using Hill notation system

        >>> str(Formula({Atom.C: 6, Atom.H: 12, Atom.O: 6}))
        'C6H12O6'
        """

        def hill_sorted_elements(values):
            def element_sort_key(pair):
                element, value = pair
                if isinstance(element, Atom):
                    return 0, element.symbol
                elif isinstance(element, Radical):
                    return 1, element.symbol
                else:
                    return 2, None

            if Atom.C in values:
                yield Atom.C, values[Atom.C]
                if Atom.H in values:
                    yield Atom.H, values[Atom.H]
                for element, value in sorted(
                        iteritems(values), key=element_sort_key):
                    if element not in (Atom.C, Atom.H):
                        yield element, value
            else:
                for element, value in sorted(
                        iteritems(values), key=element_sort_key):
                    yield element, value

        s = ''
        for element, value in hill_sorted_elements(self._values):
            def grouped(element, value):
                return '({}){}'.format(element, value if value != 1 else '')

            def nongrouped(element, value):
                return '{}{}'.format(element, value if value != 1 else '')

            if isinstance(element, Radical):
                if len(element.symbol) == 1:
                    s += nongrouped(element, value)
                else:
                    s += grouped(element, value)
            elif isinstance(element, Atom):
                s += nongrouped(element, value)
            else:
                s += grouped(element, value)
        return s

    def __repr__(self):
        return str('Formula({})').format(repr(dict(self._values)))

    def __and__(self, other):
        """Intersection of formula elements."""
        if isinstance(other, Formula):
            return Formula(self._values & other._values)
        elif isinstance(other, FormulaElement):
            return Formula(self._values & Counter([other]))
        return NotImplemented

    def __or__(self, other):
        """Merge formulas into one formula."""
        # Note: This operator corresponds to the add-operator on Counter not
        # the or-operator! The add-operator is used here (on the superclass)
        # to compose formula elements into subformulas.
        if isinstance(other, Formula):
            return Formula(self._values + other._values)
        elif isinstance(other, FormulaElement):
            return Formula(self._values + Counter([other]))
        return NotImplemented

    def __sub__(self, other):
        """Substract other formula from this formula."""
        if isinstance(other, Formula):
            return Formula(self._values - other._values)
        elif isinstance(other, FormulaElement):
            return Formula(self._values - Counter([other]))
        return NotImplemented

    def __mul__(self, other):
        """Multiply formula element by other."""
        values = {key: value*other for key, value in iteritems(self._values)}
        return Formula(values)

    def __hash__(self):
        h = hash('Formula')
        for element, value in iteritems(self._values):
            h ^= hash(element) ^ hash(value)
        return h

    def __eq__(self, other):
        return isinstance(other, Formula) and self._values == other._values

    def __ne__(self, other):
        return not self == other

    @classmethod
    def parse(cls, s):
        """Parse a formula string (e.g. C6H10O2)."""
        return _parse_formula(s)

    @classmethod
    def balance(cls, lhs, rhs):
        """Return formulas that need to be added to balance given formulas

        Given complete formulas for right side and left side of a reaction,
        calculate formulas for the missing compounds on both sides. Return
        as a left, right tuple. Formulas can be flattened before balancing
        to disregard grouping structure.
        """

        def missing(formula, other):
            for element, value in iteritems(formula._values):
                if element not in other._values:
                    yield value*element
                else:
                    delta = value - other._values[element]
                    if isinstance(delta, numbers.Number) and delta > 0:
                        yield delta*element

        return (reduce(operator.or_, missing(rhs, lhs), Formula()),
                reduce(operator.or_, missing(lhs, rhs), Formula()))


class ParseError(Exception):
    """Signals error parsing formula."""

    def __init__(self, *args, **kwargs):
        self._span = kwargs.pop('span', None)
        super(ParseError, self).__init__(*args, **kwargs)

    @property
    def indicator(self):
        if self._span is None:
            return None
        pre = ' ' * self._span[0]
        ind = '^' * max(1, self._span[1] - self._span[0])
        return pre + ind


def _parse_formula(s):
    """Parse formula string."""
    scanner = re.compile(r'''
        (\s+) |         # whitespace
        (\(|\)) |       # group
        ([A-Z][a-z]*) | # element
        (\d+) |         # number
        ([a-z]) |       # variable
        (\Z) |          # end
        (.)             # error
    ''', re.DOTALL | re.VERBOSE)

    def transform_subformula(form):
        """Extract radical if subformula is a singleton with a radical."""
        if isinstance(form, dict) and len(form) == 1:
            # A radical in a singleton subformula is interpreted as a
            # numbered radical.
            element, value = next(iteritems(form))
            if isinstance(element, Radical):
                return Radical('{}{}'.format(element.symbol, value))
        return form

    stack = []
    formula = {}
    expect_count = False

    def close(formula, count=1):
        if len(stack) == 0:
            raise ParseError('Unbalanced parenthesis group in formula')
        subformula = transform_subformula(formula)
        if isinstance(subformula, dict):
            subformula = Formula(subformula)

        formula = stack.pop()
        if subformula not in formula:
            formula[subformula] = 0
        formula[subformula] += count
        return formula

    for match in re.finditer(scanner, s):
        (whitespace, group, element, number, variable, end,
            error) = match.groups()

        if error is not None:
            raise ParseError(
                'Invalid token in formula string: {!r}'.format(match.group(0)),
                span=(match.start(), match.end()))
        elif whitespace is not None:
            continue
        elif group is not None and group == '(':
            if expect_count:
                formula = close(formula)
            stack.append(formula)
            formula = {}
            expect_count = False
        elif group is not None and group == ')':
            if expect_count:
                formula = close(formula)
            expect_count = True
        elif element is not None:
            if expect_count:
                formula = close(formula)
            stack.append(formula)
            if element in 'RX':
                formula = Radical(element)
            else:
                formula = Atom(element)
            expect_count = True
        elif number is not None and expect_count:
            formula = close(formula, int(number))
            expect_count = False
        elif variable is not None and expect_count:
            formula = close(formula, Expression(variable))
            expect_count = False
        elif end is not None:
            if expect_count:
                formula = close(formula)
        else:
            raise ParseError(
                'Invalid token in formula string: {!r}'.format(match.group(0)),
                span=(match.start(), match.end()))

    if len(stack) > 0:
        raise ParseError('Unbalanced parenthesis group in formula')

    return Formula(formula)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2020 Keith Dufault-Thompson <keitht547@uri.edu>

import unittest
import tempfile
import shutil
import os

from psamm.commands import psammotate
from psamm.datasource.native import ModelReader


class TestPsammotate(unittest.TestCase):
    def setUp(self):
        self._model_dir = tempfile.mkdtemp()
        with open(os.path.join(self._model_dir, 'model.yaml'), 'w') as f:
            f.write('\n'.join([
                '---',
                'reactions:',
                '  - id: rxn_1',
                '    equation: A[e] => B[c]',
                '    genes: gene_1',
                '  - id: rxn_2',
                '    equation: B[c] => C[c]',
                '    genes: gene_2 and gene_3',
                '  - id: rxn_3',
                '    equation: C[c] <=> D[c]',
                '    genes: gene_3 and gene_4',
                '  - id: rxn_4',
                '    equation: D[c] => E[e]',
                '    genes: gene_3 or gene_4',
                '  - id: rxn_5',
                '    equation: C[c] => D[e]',
                '    genes: (gene_3 or gene_4) and (gene_1 or gene_5)',
                '  - id: rxn_6',
                '    equation: D[c] => A[e]',
                '    genes: Biomass',
                '  - id: rxn_7',
                '    equation: E[e] => A[e]',
                '  - id: rxn_8',
                '    equation: E[e] =>'
            ]))
        self._model = ModelReader.reader_from_path(
            self._model_dir).create_model()
        self.rxn_entry_dict = {}
        for rx in self._model.reactions:
            self.rxn_entry_dict[rx.id] = rx

        with open(os.path.join(self._model_dir, 'app.tsv'), 'w') as f:
            f.write('\n'.join([
                'gene_1\tgene_a',
                'gene_2\tgene_b',
                'gene_3\tgene_c',
                'gene_4\t-',
                'gene_5\tgene_e, gene_f',
                '-\tgene_g'
            ]))

    def tearDown(self):
        shutil.rmtree(self._model_dir)

    def test_app_reader(self):
        translation_dict = psammotate.app_reader(
            open(os.path.join(self._model_dir, 'app.tsv')), 1, 0)
        self.assertTrue('-' not in translation_dict)
        self.assertTrue(translation_dict['gene_1'] == ['gene_a'])
        self.assertTrue(translation_dict['gene_2'] == ['gene_b'])
        self.assertTrue(translation_dict['gene_3'] == ['gene_c'])
        self.assertTrue(translation_dict['gene_4'] == ['-'])
        self.assertTrue(translation_dict['gene_5'] == ['gene_e', 'gene_f'])

    def test_model_loader_ignorena(self):
        translation_dict = psammotate.app_reader(
            open(os.path.join(self._model_dir, 'app.tsv')), 1, 0)
        translated_genes = psammotate.model_loader(
            self._model, True, translation_dict)
        self.assertEqual(translated_genes[self.rxn_entry_dict['rxn_1']],
                         ['gene_1', 'gene_a', True])
        self.assertEqual(translated_genes[self.rxn_entry_dict['rxn_2']],
                         ['gene_2 and gene_3', 'gene_b and gene_c', True])
        self.assertEqual(translated_genes[self.rxn_entry_dict['rxn_3']],
                         ['gene_3 and gene_4', 'gene_c and -', False])
        self.assertEqual(translated_genes[self.rxn_entry_dict['rxn_4']],
                         ['gene_3 or gene_4', 'gene_c or -', True])
        self.assertEqual(translated_genes[self.rxn_entry_dict['rxn_5']],
                         ['(gene_3 or gene_4) and (gene_1 or gene_5)',
                          '(gene_c or -) and (gene_a or (gene_e or gene_f))',
                          True])
        self.assertEqual(translated_genes[self.rxn_entry_dict['rxn_7']],
                         [None, None, False])
        self.assertEqual(translated_genes[self.rxn_entry_dict['rxn_8']],
                         [None, None, False])

    def test_model_loader(self):
        translation_dict = psammotate.app_reader(
            open(os.path.join(self._model_dir, 'app.tsv')), 1, 0)
        translated_genes = psammotate.model_loader(
            self._model, False, translation_dict)
        self.assertEqual(translated_genes[self.rxn_entry_dict['rxn_1']],
                         ['gene_1', 'gene_a', True])
        self.assertEqual(translated_genes[self.rxn_entry_dict['rxn_2']],
                         ['gene_2 and gene_3', 'gene_b and gene_c', True])
        self.assertEqual(translated_genes[self.rxn_entry_dict['rxn_3']],
                         ['gene_3 and gene_4', 'gene_c and -', False])
        self.assertEqual(translated_genes[self.rxn_entry_dict['rxn_4']],
                         ['gene_3 or gene_4', 'gene_c or -', True])
        self.assertEqual(translated_genes[self.rxn_entry_dict['rxn_5']],
                         ['(gene_3 or gene_4) and (gene_1 or gene_5)',
                          '(gene_c or -) and (gene_a or (gene_e or gene_f))',
                          True])
        self.assertEqual(translated_genes[self.rxn_entry_dict['rxn_7']],
                         [None, None, True])
        self.assertEqual(translated_genes[self.rxn_entry_dict['rxn_8']],
                         [None, None, True])

    def test_get_gene_list_and(self):
        genes = 'gene_a and gene_b'
        self.assertEqual(psammotate.get_gene_list(genes),
                         frozenset(['gene_a', 'gene_b']))

    def test_get_gene_list_or(self):
        genes = 'gene_a or gene_b'
        self.assertEqual(psammotate.get_gene_list(genes),
                         frozenset(['gene_a', 'gene_b']))

    def test_get_gene_list_nested_or(self):
        genes = 'gene_a or (gene_b or gene_c)'
        self.assertEqual(psammotate.get_gene_list(genes),
                         frozenset(['gene_a', 'gene_b', 'gene_c']))

    def test_get_gene_list_nested_and(self):
        genes = 'gene_a or (gene_b and gene_c)'
        self.assertEqual(psammotate.get_gene_list(genes),
                         frozenset(['gene_a', 'gene_b', 'gene_c']))

    def test_get_gene_list_nested_and_or(self):
        genes = '(gene_a or gene_d) or (gene_b and gene_c)'
        self.assertEqual(psammotate.get_gene_list(genes),
                         frozenset(['gene_a', 'gene_b', 'gene_c', 'gene_d']))

    def test_get_gene_list_multiple_nests(self):
        genes = '(gene_a or (gene_b and gene_c)) or gene_d'
        self.assertEqual(psammotate.get_gene_list(genes),
                         frozenset(['gene_a', 'gene_b', 'gene_c', 'gene_d']))

    def test_remove_gaps_single(self):
        genes = 'gene_a'
        self.assertEqual(psammotate.remove_gap(genes), 'gene_a')

    def test_remove_gaps_or(self):
        genes = 'gene_a or gene_b'
        self.assertEqual(psammotate.remove_gap(genes), 'gene_a or gene_b')

    def test_remove_gaps_and(self):
        genes = 'gene_a and gene_b'
        self.assertEqual(psammotate.remove_gap(genes), 'gene_a and gene_b')

    def test_remove_gaps_and_missing(self):
        genes = 'gene_a and -'
        self.assertEqual(psammotate.remove_gap(genes), 'gene_a and -')

    def test_remove_gaps_or_missing(self):
        genes = 'gene_a or -'
        self.assertEqual(psammotate.remove_gap(genes), 'gene_a')

    def test_remove_gaps_or_missing_nested(self):
        genes = 'gene_a or (- or -)'
        self.assertEqual(psammotate.remove_gap(genes), 'gene_a')

    def test_remove_gaps_and_missing_nested(self):
        genes = 'gene_a and (- or -)'
        self.assertEqual(psammotate.remove_gap(genes), 'gene_a and -')

    def test_remove_gaps_double_nested(self):
        genes = 'gene_a or (- or (- or gene_b))'
        self.assertEqual(psammotate.remove_gap(genes), 'gene_a or ((gene_b))')


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import unicode_literals

import logging
import argparse

from six import text_type

from ..command import (Command, MetabolicMixin, SolverCommandMixin,
                       FilePrefixAppendAction)
from ..gapfill import gapfill, GapFillError
from ..datasource.reaction import parse_compound
from ..gapfilling import create_extended_model

logger = logging.getLogger(__name__)


class GapFillCommand(MetabolicMixin, SolverCommandMixin, Command):
    """Run the GapFill algorithms on the model."""
    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--compound', metavar='compound', action=FilePrefixAppendAction,
            type=parse_compound, default=[],
            help='Select compounds to try to unblock')
        parser.add_argument(
            '--penalty', metavar='file', type=argparse.FileType('r'),
            help='List of penalty scores for database reactions')
        parser.add_argument(
            '--db-penalty', metavar='penalty', type=float,
            help='Default penalty for database reactions')
        parser.add_argument(
            '--tp-penalty', metavar='penalty', type=float,
            help='Default penalty for transport reactions')
        parser.add_argument(
            '--ex-penalty', metavar='penalty', type=float,
            help='Default penalty for exchange reactions')
        parser.add_argument(
            '--epsilon', type=float, default=1e-5,
            help='Threshold for reaction flux')
        parser.add_argument(
            '--no-implicit-sinks', action='store_true',
            help='Do not include implicit sinks when gap-filling')
        parser.add_argument(
            '--allow-bounds-expansion', action='store_true',
            help=('Allow GapFill to propose expansion of flux bounds. This'
                  ' includes turning irreversible reactions reversible.'))
        super(GapFillCommand, cls).init_parser(parser)

    def run(self):
        """Run GapFill command"""

        # Load compound information
        def compound_name(id):
            if id not in self._model.compounds:
                return id
            return self._model.compounds[id].properties.get('name', id)

        # Calculate penalty if penalty file exists
        penalties = {}
        if self._args.penalty is not None:
            for line in self._args.penalty:
                line, _, comment = line.partition('#')
                line = line.strip()
                if line == '':
                    continue
                rxnid, penalty = line.split(None, 1)
                penalties[rxnid] = float(penalty)

        core = set(self._mm.reactions)

        solver = self._get_solver(integer=True)
        default_comp = self._model.default_compartment
        epsilon = self._args.epsilon
        v_max = float(self._model.default_flux_limit)

        blocked = set()
        for compound in self._args.compound:
            if compound.compartment is None:
                compound = compound.in_compartment(default_comp)
            blocked.add(compound)

        if len(blocked) > 0:
            logger.info('Unblocking compounds: {}...'.format(
                ', '.join(text_type(c) for c in sorted(blocked))))
        else:
            logger.info(
                'Unblocking all compounds in model. Use --compound option to'
                ' unblock specific compounds.')
            blocked = set(self._mm.compounds)

        exclude = set()
        if self._model.biomass_reaction is not None:
            exclude.add(self._model.biomass_reaction)

        # Add exchange and transport reactions to database
        model_complete, weights = create_extended_model(
            self._model,
            db_penalty=self._args.db_penalty,
            ex_penalty=self._args.ex_penalty,
            tp_penalty=self._args.tp_penalty,
            penalties=penalties)

        implicit_sinks = not self._args.no_implicit_sinks

        logger.info('Searching for reactions to fill gaps')
        try:
            added_reactions, no_bounds_reactions = gapfill(
                model_complete, core, blocked, exclude, solver=solver,
                epsilon=epsilon, v_max=v_max, weights=weights,
                implicit_sinks=implicit_sinks,
                allow_bounds_expansion=self._args.allow_bounds_expansion)
        except GapFillError as e:
            self._log_epsilon_and_fail(epsilon, e)

        for reaction_id in sorted(self._mm.reactions):
            rx = self._mm.get_reaction(reaction_id)
            rxt = rx.translated_compounds(compound_name)
            print('{}\t{}\t{}\t{}'.format(reaction_id, 'Model', 0, rxt))

        for rxnid in sorted(added_reactions):
            rx = model_complete.get_reaction(rxnid)
            rxt = rx.translated_compounds(compound_name)
            print('{}\t{}\t{}\t{}'.format(
                rxnid, 'Add', weights.get(rxnid, 1), rxt))

        for rxnid in sorted(no_bounds_reactions):
            rx = model_complete.get_reaction(rxnid)
            rxt = rx.translated_compounds(compound_name)
            print('{}\t{}\t{}\t{}'.format(
                rxnid, 'Remove bounds', weights.get(rxnid, 1), rxt))

    def _log_epsilon_and_fail(self, epsilon, exc):
        msg = ('Finding blocked compounds failed with epsilon set to {}. Try'
               ' lowering the epsilon value to reduce artifical constraints on'
               ' the model.'.format(epsilon))
        self.fail(msg, exc)



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018-2020  Ke Zhang <kzhang@my.uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@uri.edu>
# Copyright 2020-2020  Elysha Sameth <esameth@my.uri.edu>

from __future__ import unicode_literals

import unittest
import os
import tempfile

from psamm.formula import Formula, Atom
from psamm.commands import vis
from psamm.reaction import Compound, Reaction, Direction
from collections import defaultdict
from psamm.datasource.reaction import parse_reaction, parse_compound
from psamm.datasource import entry
from psamm import graph
from psamm.datasource.native import NativeModel, ReactionEntry, CompoundEntry


# 2019-08-30 new test cases for vis
class TestMakeSubset(unittest.TestCase):
    def setUp(self):
        native_model = NativeModel()
        native_model.reactions.add_entry(ReactionEntry({
            'id': 'rxn1', 'equation': parse_reaction(
                'fum[c] + h2o[c] <=> mal_L[c]')}))
        native_model.compounds.add_entry(CompoundEntry({
            'id': 'fum', 'formula': parse_compound('C4H2O4', 'c')}))
        native_model.compounds.add_entry(CompoundEntry({
            'id': 'h2o', 'formula': parse_compound('H2O', 'c')}))
        native_model.compounds.add_entry(CompoundEntry(
            {'id': 'mal_L', 'formula': parse_compound('C4H4O5', 'c')}))
        native_model.reactions.add_entry(ReactionEntry({
            'id': 'rxn2', 'equation': parse_reaction(
                'q8[c] + succ[c] => fum[c] + q8h2[c]')}))
        native_model.compounds.add_entry(CompoundEntry(
            {'id': 'q8', 'formula': parse_compound('C49H74O4', 'c')}))
        native_model.compounds.add_entry(CompoundEntry({
            'id': 'q8h2', 'formula': parse_compound('C49H76O4', 'c')}))
        native_model.compounds.add_entry(CompoundEntry({
            'id': 'succ', 'formula': parse_compound('C4H4O4', 'c')}))
        native_model.reactions.add_entry(ReactionEntry({
            'id': 'rxn3', 'equation': parse_reaction(
                'h2o[c] <=> h2o[e]')}))
        native_model.compounds.add_entry(CompoundEntry({
            'id': 'h2o', 'formula': parse_compound('H2O', 'c')}))
        self.native = native_model
        self.mm = native_model.create_metabolic_model()
        self.sub = None
        self.exclude = []

    def test1_no_subset(self):
        subset = vis.rxnset_for_vis(self.mm, self.sub, self.exclude)
        self.assertEqual(subset, set(['rxn1', 'rxn2', 'rxn3']))

    def test2_subset_reactions(self):
        path = os.path.join(tempfile.mkdtemp(), 'subset')
        with open(path, 'w') as f:
            f.write('{}\n{}'.format('rxn1', 'rxn2'))
        sub_file = open(path, 'r')
        subset = vis.rxnset_for_vis(self.mm, sub_file, self.exclude)
        sub_file.close()
        self.assertEqual(subset, set(['rxn1', 'rxn2']))

    def test3_subset_one_compound1(self):
        path = os.path.join(tempfile.mkdtemp(), 'subset')
        with open(path, 'w') as f:
            f.write('{}'.format('mal_L[c]'))
        sub_file = open(path, 'r')
        subset = vis.rxnset_for_vis(self.mm, sub_file, self.exclude)
        sub_file.close()
        self.assertEqual(subset, set(['rxn1']))

    def test4_subset_one_compound2(self):
        path = os.path.join(tempfile.mkdtemp(), 'subset')
        with open(path, 'w') as f:
            f.write('{}'.format('fum[c]'))
        sub_file = open(path, 'r')
        subset = vis.rxnset_for_vis(self.mm, sub_file, self.exclude)
        sub_file.close()
        self.assertEqual(subset, set(['rxn1', 'rxn2']))

    def test5_subset_compound3(self):
        path = os.path.join(tempfile.mkdtemp(), 'subset')
        with open(path, 'w') as f:
            f.write('{}'.format('h2o[c]'))
        sub_file = open(path, 'r')
        subset = vis.rxnset_for_vis(self.mm, sub_file, self.exclude)
        sub_file.close()
        self.assertEqual(subset, set(['rxn1', 'rxn3']))

    def test6_mixed_subset(self):
        path = os.path.join(tempfile.mkdtemp(), 'subset')
        with open(path, 'w') as f:
            f.write('{}\n{}'.format('h2o[c]', 'rxn1'))
        sub_file = open(path, 'r')
        self.assertRaises(ValueError, vis.rxnset_for_vis,
                          self.mm, sub_file, self.exclude)

    def test7_not_in_model(self):
        path = os.path.join(tempfile.mkdtemp(), 'subset')
        with open(path, 'w') as f:
            f.write('{}'.format('h3o'))
        sub_file = open(path, 'r')
        self.assertRaises(ValueError, vis.rxnset_for_vis,
                          self.mm, sub_file, self.exclude)

    def test8_exclude(self):
        exclude = ['rxn3']
        subset = vis.rxnset_for_vis(self.mm, self.sub, exclude)
        self.assertEqual(subset, set(['rxn1', 'rxn2']))

    def test9_exclude_mult(self):
        exclude = ['rxn1', 'rxn3']
        subset = vis.rxnset_for_vis(self.mm, self.sub, exclude)
        self.assertEqual(subset, set(['rxn2']))


class TestAddNodeProps(unittest.TestCase):
    def setUp(self):
        self.cpd_a = CompoundEntry({
            'id': 'A', 'name': 'compound A', 'formula': 'XXXX', 'charge': 0})
        self.cpd_b = CompoundEntry({
            'id': 'B', 'name': 'compound B', 'formula': 'YYYY', 'charge': 0})
        self.cpd_c = CompoundEntry({
            'id': 'C', 'name': 'compound C', 'formula': 'ZZZZ', 'charge': 0})
        self.cpd_d = CompoundEntry({
            'id': 'D', 'name': 'compound D', 'formula': 'MMMM', 'charge': 0})
        self.cpd_e = CompoundEntry({
            'id': 'E', 'name': 'compound E', 'formula': 'NNNN', 'charge': 0})

        # new rxn used to test: rxn1: A => B; rxn2: A => B; rxn3: C <=> B + D
        self.rxn1 = entry.DictReactionEntry({
            'id': 'rxn1', 'name': 'Reaction 1', 'genes': 'gene1 and gene2',
            'equation': Reaction(
                Direction.Forward, {Compound('A', 'c'): -1,
                                    Compound('B', 'c'): 1}),
            })
        self.rxn2 = entry.DictReactionEntry({
            'id': 'rxn2', 'name': 'Reaction 2', 'genes': 'gene3',
            'equation': Reaction(Direction.Forward, {
                Compound('A', 'c'): -1, Compound('B', 'c'): 1})
            })
        self.rxn3 = entry.DictReactionEntry({
            'id': 'rxn3', 'name': 'Reaction 3',
            'equation': Reaction(Direction.Both, {
                Compound('C', 'c'): -1, Compound('B', 'c'): 1,
                Compound('D', 'c'): 1})
        })

        self.node_a = graph.Node({
            'id': 'A[c]', 'entry': [self.cpd_a],
            'type': 'cpd', 'compartment': 'c'})
        self.node_b = graph.Node({
            'id': 'B[c]', 'entry': [self.cpd_b],
            'type': 'cpd', 'compartment': 'c'})
        self.node_c = graph.Node({
            'id': 'C[c]', 'entry': [self.cpd_c],
            'type': 'cpd', 'compartment': 'c'})
        self.node_d = graph.Node({
            'id': 'D[c]', 'entry': [self.cpd_d],
            'type': 'cpd', 'compartment': 'c'})
        self.node_ab = graph.Node({
            'id': 'rxn1_1, rxn2_1', 'entry': [self.rxn1, self.rxn2],
            'compartment': 'c', 'type': 'rxn'})
        self.node_cb = graph.Node({
            'id': 'rxn3_1', 'entry': [self.rxn3], 'compartment': 'c',
            'type': 'rxn'})

        self.edge1 = graph.Edge(self.node_a, self.node_ab, {'dir': 'forward'})
        self.edge2 = graph.Edge(self.node_ab, self.node_b, {'dir': 'forward'})
        self.edge3 = graph.Edge(self.node_c, self.node_cb, {'dir': 'both'})
        self.edge4 = graph.Edge(self.node_cb, self.node_b, {'dir': 'both'})
        self.edge5 = graph.Edge(self.node_cb, self.node_d, {'dir': 'both'})

        self.g = graph.Graph()
        self.node_list = [self.node_a, self.node_b, self.node_c, self.node_d,
                          self.node_ab, self.node_cb]
        for node in self.node_list:
            self.g.add_node(node)
        self.edge_list = [self.edge1, self.edge2,
                          self.edge3, self.edge4, self.edge5]
        for edge in self.edge_list:
            self.g.add_edge(edge)

        self.recolor_dict = {'A': '#f4fc55', 'rxn1': '#ef70de',
                             'rxn3': '#64b3f4'}

    def test1_DefaultRecolor(self):
        g1 = vis.add_node_props(self.g, {})
        node_a = graph.Node({
            'id': 'A[c]', 'entry': [self.cpd_a],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf'})
        node_b = graph.Node({
            'id': 'B[c]', 'entry': [self.cpd_b],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf'})
        node_c = graph.Node({
            'id': 'C[c]', 'entry': [self.cpd_c],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf'})
        node_d = graph.Node({
            'id': 'D[c]', 'entry': [self.cpd_d],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf'})
        node_ab = graph.Node({
            'id': 'rxn1_1, rxn2_1', 'entry': [self.rxn1, self.rxn2],
            'compartment': 'c', 'type': 'rxn', 'style': 'filled',
            'shape': 'box', 'fillcolor': '#c9fccd'})
        node_cb = graph.Node({
            'id': 'rxn3_1', 'entry': [self.rxn3],
            'compartment': 'c', 'type': 'rxn',
            'style': 'filled', 'shape': 'box',
            'fillcolor': '#c9fccd'})
        node_list = [node_a, node_b, node_c, node_d, node_ab, node_cb]
        self.assertTrue(all(i in node_list for i in g1.nodes))
        self.assertTrue(all(i in g1.nodes for i in node_list))

    def test2_Recolor(self):
        g2 = vis.add_node_props(self.g, self.recolor_dict)
        node_a = graph.Node({
            'id': 'A[c]', 'entry': [self.cpd_a],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#f4fc55'})
        node_b = graph.Node({
            'id': 'B[c]', 'entry': [self.cpd_b],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf'})
        node_c = graph.Node({
            'id': 'C[c]', 'entry': [self.cpd_c],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf'})
        node_d = graph.Node({
            'id': 'D[c]', 'entry': [self.cpd_d],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf'})
        node_ab = graph.Node({
            'id': 'rxn1_1, rxn2_1', 'entry': [self.rxn1, self.rxn2],
            'compartment': 'c', 'type': 'rxn', 'style': 'filled',
            'shape': 'box', 'fillcolor': '#c9fccd'})
        node_cb = graph.Node({
            'id': 'rxn3_1', 'entry': [self.rxn3],
            'compartment': 'c', 'type': 'rxn',
            'style': 'filled', 'shape': 'box',
            'fillcolor': '#64b3f4'})
        node_list = [node_a, node_b, node_c, node_d, node_ab, node_cb]
        self.assertTrue(all(i in node_list for i in g2.nodes))
        self.assertTrue(all(i in g2.nodes for i in node_list))

    def test3_ConflictRecolorDict(self):
        recolor_dict = {'A': '#f4fc55', 'rxn1': '#ef70de', 'rxn2': '#64b3f4'}
        g3 = vis.add_node_props(self.g, recolor_dict)
        node_a = graph.Node({
            'id': 'A[c]', 'entry': [self.cpd_a],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#f4fc55'})
        node_b = graph.Node({
            'id': 'B[c]', 'entry': [self.cpd_b],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf'})
        node_c = graph.Node({
            'id': 'C[c]', 'entry': [self.cpd_c],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf'})
        node_d = graph.Node({
            'id': 'D[c]', 'entry': [self.cpd_d],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf'})
        node_ab = graph.Node({
            'id': 'rxn1_1, rxn2_1', 'entry': [self.rxn1, self.rxn2],
            'compartment': 'c', 'type': 'rxn', 'style': 'filled',
            'shape': 'box', 'fillcolor': '#c9fccd'})
        node_cb = graph.Node({
            'id': 'rxn3_1', 'entry': [self.rxn3],
            'compartment': 'c', 'type': 'rxn',
            'style': 'filled', 'shape': 'box',
            'fillcolor': '#c9fccd'})
        node_list = [node_a, node_b, node_c, node_d, node_ab, node_cb]
        self.assertTrue(all(i in node_list for i in g3.nodes))
        self.assertTrue(all(i in g3.nodes for i in node_list))


class TestAddNodeLabel(unittest.TestCase):
    def setUp(self):
        self.cpd_a = CompoundEntry({
            'id': 'A', 'name': 'compound A', 'formula': 'XXXX', 'charge': 0})
        self.cpd_b = CompoundEntry({
            'id': 'B', 'name': 'compound B', 'formula': 'YYYY', 'charge': 0})
        self.cpd_c = CompoundEntry({
            'id': 'C', 'name': 'compound C', 'formula': 'ZZZZ', 'charge': 0})
        self.cpd_d = CompoundEntry({
            'id': 'D', 'name': 'compound D', 'formula': 'MMMM', 'charge': 0})
        # self.cpd_e = CompoundEntry({
        #     'id': 'E', 'name': 'compound E', 'formula': 'NNNN', 'charge': 0})

        # new rxn used to test: rxn1: A => B; rxn2: A => B; rxn3: C <=> B + D
        self.rxn1 = entry.DictReactionEntry({
            'id': 'rxn1', 'name': 'Reaction 1', 'genes': 'gene1 and gene2',
            'equation': Reaction(
                Direction.Forward, {Compound('A', 'c'): -1,
                                    Compound('B', 'c'): 1}),
        })
        self.rxn2 = entry.DictReactionEntry({
            'id': 'rxn2', 'name': 'Reaction 2',
            'equation': Reaction(Direction.Forward, {
                Compound('A', 'c'): -1, Compound('B', 'c'): 1})
        })
        self.rxn3 = entry.DictReactionEntry({
            'id': 'rxn3', 'name': 'Reaction 3', 'genes': 'gene3',
            'equation': Reaction(Direction.Both, {
                Compound('C', 'c'): -1, Compound('B', 'c'): 1,
                Compound('D', 'c'): 1})
        })

        self.node_a = graph.Node({
            'id': 'A[c]', 'entry': [self.cpd_a],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf'})
        self.node_b = graph.Node({
            'id': 'B[c]', 'entry': [self.cpd_b],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf'})
        self.node_c = graph.Node({
            'id': 'C[c]', 'entry': [self.cpd_c],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf'})
        self.node_d = graph.Node({
            'id': 'D[c]', 'entry': [self.cpd_d],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf'})
        self.node_ab = graph.Node({
            'id': 'rxn1_1, rxn2_1', 'entry': [self.rxn1, self.rxn2],
            'compartment': 'c', 'type': 'rxn', 'style': 'filled',
            'shape': 'box', 'fillcolor': '#c9fccd'})
        self.node_cb = graph.Node({
            'id': 'rxn3_1', 'entry': [self.rxn3],
            'compartment': 'c', 'type': 'rxn',
            'style': 'filled', 'shape': 'box',
            'fillcolor': '#c9fccd'})
        self.edge1 = graph.Edge(self.node_a, self.node_ab, {'dir': 'forward'})
        self.edge2 = graph.Edge(self.node_ab, self.node_b, {'dir': 'forward'})
        self.edge3 = graph.Edge(self.node_c, self.node_cb, {'dir': 'both'})
        self.edge4 = graph.Edge(self.node_cb, self.node_b, {'dir': 'both'})
        self.edge5 = graph.Edge(self.node_cb, self.node_d, {'dir': 'both'})

        self.g = graph.Graph()
        node_list = [self.node_a, self.node_b,
                     self.node_c, self.node_d, self.node_ab, self.node_cb]
        # node_list = [a, b, c, d, ab, cb]
        for node in node_list:
            self.g.add_node(node)
        edge_list = [self.edge1, self.edge2,
                     self.edge3, self.edge4, self.edge5]
        for edge in edge_list:
            self.g.add_edge(edge)

    def test1_detail_default(self):
        """test vis codes when both --cpd-detail and --rxn-detail are None"""
        g1 = vis.add_node_label(self.g, None, None)
        # self.node_a.props['label'] = 'A[c]'
        # self.node_b.props['label'] = 'B[c]'
        # self.node_c.props['label'] = 'C[c]'
        # self.node_d.props['label'] = 'D[c]'
        # self.node_ab.props['label'] = 'rxn1\nrxn2'
        # self.node_cb.props['label'] = 'rxn3\nReaction 3'
        # node_list_1 = [self.node_a, self.node_b, self.node_c, self.node_d,
        #                self.node_ab, self.node_cb]

        a = graph.Node({
            'id': 'A[c]', 'entry': [self.cpd_a],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'A[c]'})
        b = graph.Node({
            'id': 'B[c]', 'entry': [self.cpd_b],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'B[c]'})
        c = graph.Node({
            'id': 'C[c]', 'entry': [self.cpd_c],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'C[c]'})
        d = graph.Node({
            'id': 'D[c]', 'entry': [self.cpd_d],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'D[c]'})
        ab = graph.Node({
            'id': 'rxn1_1, rxn2_1', 'entry': [self.rxn1, self.rxn2],
            'compartment': 'c', 'type': 'rxn', 'style': 'filled',
            'shape': 'box', 'fillcolor': '#c9fccd', 'label': 'rxn1\nrxn2'})
        cb = graph.Node({
            'id': 'rxn3_1', 'entry': [self.rxn3],
            'compartment': 'c', 'type': 'rxn',
            'style': 'filled', 'shape': 'box',
            'fillcolor': '#c9fccd', 'label': 'rxn3'})
        node_list_1 = [a, b, c, d, ab, cb]
        # for i in g1.nodes:
        #     print('in g1:', i.props)
        #     if i not in node_list_1:
        #         print('in g1 but not in node_list:', i)
        # for j in node_list_1:
        #     print('in node list:', j.props)
        self.assertTrue(all(i in node_list_1 for i in g1.nodes))
        self.assertTrue(all(i in g1.nodes for i in node_list_1))

    def test2_detail_id_name(self):
        """test vis codes when both --cpd-detail
        and --rxn-detail are ['id', 'name']"""
        g2 = vis.add_node_label(self.g, [['id', 'name']], [['id', 'name']])
        # self.node_a.props['label'] = 'A[c]\ncompound A'
        # self.node_b.props['label'] = 'B[c]\ncompound B'
        # self.node_c.props['label'] = 'C[c]\ncompound C'
        # self.node_d.props['label'] = 'D[c]\ncompound D'
        # self.node_ab.props['label'] = 'rxn1\nrxn2'
        # self.node_cb.props['label'] = 'rxn3\nReaction 3'
        # node_list_2 = [self.node_a, self.node_b, self.node_c, self.node_d,
        #                self.node_ab, self.node_cb]

        a = graph.Node({
            'id': 'A[c]', 'entry': [self.cpd_a],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'A[c]\ncompound A'})
        b = graph.Node({
            'id': 'B[c]', 'entry': [self.cpd_b],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'B[c]\ncompound B'})
        c = graph.Node({
            'id': 'C[c]', 'entry': [self.cpd_c],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'C[c]\ncompound C'})
        d = graph.Node({
            'id': 'D[c]', 'entry': [self.cpd_d],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'D[c]\ncompound D'})
        ab = graph.Node({
            'id': 'rxn1_1, rxn2_1', 'entry': [self.rxn1, self.rxn2],
            'compartment': 'c', 'type': 'rxn', 'style': 'filled',
            'shape': 'box', 'fillcolor': '#c9fccd', 'label': 'rxn1\nrxn2'})
        cb = graph.Node({
            'id': 'rxn3_1', 'entry': [self.rxn3],
            'compartment': 'c', 'type': 'rxn',
            'style': 'filled', 'shape': 'box',
            'fillcolor': '#c9fccd', 'label': 'rxn3\nReaction 3'})
        node_list_2 = [a, b, c, d, ab, cb]
        for i in g2.nodes:
            if i not in node_list_2:
                print('not in node_list :', i, i.props)
        self.assertTrue(all(i in node_list_2 for i in g2.nodes))
        self.assertTrue(all(i in g2.nodes for i in node_list_2))

    def test3_id_formula_genes(self):
        """test vis codes when --cpd-detail and
        --rxn-detail are ['id', 'formula', 'genes']"""
        g3 = vis.add_node_label(self.g, [['id', 'formula', 'genes']],
                                [['id', 'formula', 'genes']])
        # self.node_a.props['label'] = 'A[c]\ncompound A'
        # self.node_b.props['label'] = 'B[c]\ncompound B'
        # self.node_c.props['label'] = 'C[c]\ncompound C'
        # self.node_d.props['label'] = 'D[c]\ncompound D'
        # self.node_ab.props['label'] = 'rxn1\nrxn2'
        # self.node_cb.props['label'] = 'rxn3\nReaction 3'
        # node_list_2 = [self.node_a, self.node_b, self.node_c, self.node_d,
        #                self.node_ab, self.node_cb]

        a = graph.Node({
            'id': 'A[c]', 'entry': [self.cpd_a],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'A[c]\nXXXX'})
        b = graph.Node({
            'id': 'B[c]', 'entry': [self.cpd_b],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'B[c]\nYYYY'})
        c = graph.Node({
            'id': 'C[c]', 'entry': [self.cpd_c],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'C[c]\nZZZZ'})
        d = graph.Node({
            'id': 'D[c]', 'entry': [self.cpd_d],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'D[c]\nMMMM'})
        ab = graph.Node({
            'id': 'rxn1_1, rxn2_1', 'entry': [self.rxn1, self.rxn2],
            'compartment': 'c', 'type': 'rxn', 'style': 'filled',
            'shape': 'box', 'fillcolor': '#c9fccd', 'label': 'rxn1\nrxn2'})
        cb = graph.Node({
            'id': 'rxn3_1', 'entry': [self.rxn3],
            'compartment': 'c', 'type': 'rxn',
            'style': 'filled', 'shape': 'box',
            'fillcolor': '#c9fccd', 'label': 'rxn3\ngene3'})
        node_list_3 = [a, b, c, d, ab, cb]
        self.assertTrue(all(i in node_list_3 for i in g3.nodes))
        self.assertTrue(all(i in g3.nodes for i in node_list_3))

    def test4_noID(self):
        """test vis codes when 'id' is not in --cpd-detail and --rxn-detail"""
        g4 = vis.add_node_label(self.g, [['name', 'genes']],
                                [['name', 'formula', 'genes']])
        # self.node_a.props['label'] = 'A[c]\ncompound A'
        # self.node_b.props['label'] = 'B[c]\ncompound B'
        # self.node_c.props['label'] = 'C[c]\ncompound C'
        # self.node_d.props['label'] = 'D[c]\ncompound D'
        # self.node_ab.props['label'] = 'rxn1\nrxn2'
        # self.node_cb.props['label'] = 'rxn3\nReaction 3'
        # node_list_2 = [self.node_a, self.node_b, self.node_c, self.node_d,
        #                self.node_ab, self.node_cb]

        a = graph.Node({
            'id': 'A[c]', 'entry': [self.cpd_a],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'compound A'})
        b = graph.Node({
            'id': 'B[c]', 'entry': [self.cpd_b],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'compound B'})
        c = graph.Node({
            'id': 'C[c]', 'entry': [self.cpd_c],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'compound C'})
        d = graph.Node({
            'id': 'D[c]', 'entry': [self.cpd_d],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'compound D'})
        ab = graph.Node({
            'id': 'rxn1_1, rxn2_1', 'entry': [self.rxn1, self.rxn2],
            'compartment': 'c', 'type': 'rxn', 'style': 'filled',
            'shape': 'box', 'fillcolor': '#c9fccd', 'label': 'rxn1\nrxn2'})
        cb = graph.Node({
            'id': 'rxn3_1', 'entry': [self.rxn3],
            'compartment': 'c', 'type': 'rxn',
            'style': 'filled', 'shape': 'box',
            'fillcolor': '#c9fccd', 'label': 'Reaction 3\ngene3'})
        node_list_4 = [a, b, c, d, ab, cb]
        self.assertTrue(all(i in node_list_4 for i in g4.nodes))
        self.assertTrue(all(i in g4.nodes for i in node_list_4))

    def test5_(self):
        """test vis codes when obly --cpd-detail is given """
        g5 = vis.add_node_label(self.g, [['name', 'genes']], None)
        # self.node_a.props['label'] = 'A[c]\ncompound A'
        # self.node_b.props['label'] = 'B[c]\ncompound B'
        # self.node_c.props['label'] = 'C[c]\ncompound C'
        # self.node_d.props['label'] = 'D[c]\ncompound D'
        # self.node_ab.props['label'] = 'rxn1\nrxn2'
        # self.node_cb.props['label'] = 'rxn3\nReaction 3'
        # node_list_2 = [self.node_a, self.node_b, self.node_c, self.node_d,
        #                self.node_ab, self.node_cb]

        a = graph.Node({
            'id': 'A[c]', 'entry': [self.cpd_a],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'compound A'})
        b = graph.Node({
            'id': 'B[c]', 'entry': [self.cpd_b],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'compound B'})
        c = graph.Node({
            'id': 'C[c]', 'entry': [self.cpd_c],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'compound C'})
        d = graph.Node({
            'id': 'D[c]', 'entry': [self.cpd_d],
            'type': 'cpd', 'compartment': 'c',
            'style': 'filled', 'shape': 'ellipse',
            'fillcolor': '#ffd8bf', 'label': 'compound D'})
        ab = graph.Node({
            'id': 'rxn1_1, rxn2_1', 'entry': [self.rxn1, self.rxn2],
            'compartment': 'c', 'type': 'rxn', 'style': 'filled',
            'shape': 'box', 'fillcolor': '#c9fccd', 'label': 'rxn1\nrxn2'})
        cb = graph.Node({
            'id': 'rxn3_1', 'entry': [self.rxn3],
            'compartment': 'c', 'type': 'rxn',
            'style': 'filled', 'shape': 'box',
            'fillcolor': '#c9fccd', 'label': 'rxn3'})
        node_list_5 = [a, b, c, d, ab, cb]
        self.assertTrue(all(i in node_list_5 for i in g5.nodes))
        self.assertTrue(all(i in g5.nodes for i in node_list_5))


class TestAddBiomassRnxs(unittest.TestCase):
    def setUp(self):
        cpd_a = CompoundEntry({
            'id': 'A', 'name': 'compound A', 'formula': 'XXXX', 'charge': 0})
        cpd_b = CompoundEntry({
            'id': 'B', 'name': 'compound B', 'formula': 'YYYY', 'charge': 0})
        cpd_c = CompoundEntry({
            'id': 'C', 'name': 'compound C', 'formula': 'ZZZZ', 'charge': 0})
        cpd_d = CompoundEntry({
            'id': 'D', 'name': 'compound D', 'formula': 'MMMM', 'charge': 0})
        rxn1 = entry.DictReactionEntry({
            'id': 'rxn1', 'name': 'Reaction 1', 'genes': 'gene1 and gene2',
            'equation': Reaction(
                Direction.Both, {Compound('A', 'c'): -1,
                                 Compound('C', 'c'): 1}),
        })

        self.node_a = graph.Node({
            'id': 'A[c]', 'shape': 'ellipse', 'style': 'filled', 'type':
                'cpd', 'entry': [cpd_a], 'compartment': 'c',
            'fillcolor': '#ffd8bf'})
        self.node_b = graph.Node({
            'id': 'B[c]', 'shape': 'ellipse', 'style': 'filled', 'type':
                'cpd', 'entry': [cpd_b], 'compartment': 'c',
            'fillcolor': '#ffd8bf'})
        self.node_c = graph.Node({
            'id': 'C[c]', 'shape': 'ellipse', 'style': 'filled', 'type':
                'cpd', 'entry': [cpd_c], 'compartment': 'c',
            'fillcolor': '#ffd8bf'})
        self.node_d = graph.Node({
            'id': 'D[c]', 'shape': 'ellipse', 'style': 'filled', 'type':
                'cpd', 'entry': [cpd_d], 'compartment': 'c',
            'fillcolor': '#ffd8bf'})
        self.node_rxn1 = graph.Node({
            'id': 'rxn1_1', 'shape': 'box', 'style': 'filled', 'type': 'rxn',
            'entry': [rxn1], 'compartment': 'c', 'fillcolor': '#c9fccd'})
        self.edge_A_rxn = graph.Edge(self.node_a, self.node_rxn1,
                                     {'dir': 'both'})
        self.edge_C_rxn = graph.Edge(self.node_rxn1, self.node_c,
                                     {'dir': 'both'})
        self.g = graph.Graph()
        for node in [self.node_a, self.node_b,
                     self.node_c, self.node_d, self.node_rxn1]:
            self.g.add_node(node)
        self.g.add_edge(self.edge_A_rxn)
        self.g.add_edge(self.edge_C_rxn)

        self.biorxn = ReactionEntry({
            'id': 'test_bio', 'name': 'biomass_equatin', 'equation':
                parse_reaction('A[c] + B[c] => D[c]')})

    def test1_addBiorxn_default(self):
        g1 = vis.add_biomass_rxns(self.g, self.biorxn)
        bio_a = graph.Node({
            'id': 'test_bio_1', 'entry': [self.biorxn], 'shape': 'box',
            'style': 'filled', 'label': 'test_bio', 'type': 'bio_rxn',
            'fillcolor': '#b3fcb8', 'compartment': 'c'})
        bio_b = graph.Node({
            'id': 'test_bio_2', 'entry': [self.biorxn], 'shape': 'box',
            'style': 'filled', 'label': 'test_bio', 'type': 'bio_rxn',
            'fillcolor': '#b3fcb8', 'compartment': 'c'})
        bio_d = graph.Node({
            'id': 'test_bio_3', 'entry': [self.biorxn], 'shape': 'box',
            'style': 'filled', 'label': 'test_bio', 'type': 'bio_rxn',
            'fillcolor': '#b3fcb8', 'compartment': 'c'})
        edge_bio_a = graph.Edge(self.node_a, bio_a, {'dir': 'forward'})
        edge_bio_b = graph.Edge(self.node_b, bio_b, {'dir': 'forward'})
        edge_bio_d = graph.Edge(bio_d, self.node_d, {'dir': 'forward'})

        self.assertTrue(all(i in [self.node_a, self.node_b,
                                  self.node_c, self.node_d,
                                  self.node_rxn1, bio_a, bio_b, bio_d]
                            for i in g1.nodes))
        self.assertTrue(all(i in g1.nodes for i in [
            self.node_a, self.node_b, self.node_c, self.node_d, self.node_rxn1,
            bio_a, bio_b, bio_d]))

        self.assertTrue(all(i in [self.edge_A_rxn, self.edge_C_rxn, edge_bio_a,
                                  edge_bio_b, edge_bio_d] for i in g1.edges))
        self.assertTrue(all(i in g1.edges for i in [
            self.edge_A_rxn, self.edge_C_rxn,
            edge_bio_a, edge_bio_b, edge_bio_d]))

    def test2_TwoCpdsRight(self):
        biorxn = ReactionEntry({
            'id': 'test_bio', 'name': 'biomass_equatin',
            'equation': parse_reaction('A[c] + B[c] => C[c] + D[c]')})
        g2 = vis.add_biomass_rxns(self.g, biorxn)
        bio_a = graph.Node({
            'id': 'test_bio_1', 'entry': [biorxn], 'shape': 'box',
            'style': 'filled', 'label': 'test_bio', 'type': 'bio_rxn',
            'fillcolor': '#b3fcb8', 'compartment': 'c'})
        bio_b = graph.Node({
            'id': 'test_bio_2', 'entry': [biorxn], 'shape': 'box',
            'style': 'filled', 'label': 'test_bio', 'type': 'bio_rxn',
            'fillcolor': '#b3fcb8', 'compartment': 'c'})
        bio_c = graph.Node({
            'id': 'test_bio_3', 'entry': [biorxn], 'shape': 'box',
            'style': 'filled', 'label': 'test_bio', 'type': 'bio_rxn',
            'fillcolor': '#b3fcb8', 'compartment': 'c'})
        bio_d = graph.Node({
            'id': 'test_bio_4', 'entry': [biorxn], 'shape': 'box',
            'style': 'filled', 'label': 'test_bio', 'type': 'bio_rxn',
            'fillcolor': '#b3fcb8', 'compartment': 'c'})
        edge_bio_a = graph.Edge(self.node_a, bio_a, {'dir': 'forward'})
        edge_bio_b = graph.Edge(self.node_b, bio_b, {'dir': 'forward'})
        edge_bio_c = graph.Edge(bio_c, self.node_c, {'dir': 'forward'})
        edge_bio_d = graph.Edge(bio_d, self.node_d, {'dir': 'forward'})

        final_nodes = [self.node_a, self.node_b, self.node_c, self.node_d,
                       self.node_rxn1, bio_a, bio_b, bio_c, bio_d]
        final_edges = [self.edge_A_rxn, self.edge_C_rxn, edge_bio_a,
                       edge_bio_b, edge_bio_c, edge_bio_d]
        self.assertTrue(all(i in final_nodes for i in g2.nodes))
        self.assertTrue(all(i in g2.nodes for i in final_nodes))
        self.assertTrue(all(i in final_edges for i in g2.edges))
        self.assertTrue(all(i in g2.edges for i in final_edges))

    def test3_EmptyRight(self):
        biorxn = ReactionEntry({
            'id': 'test_bio', 'name': 'biomass_equatin',
            'equation': parse_reaction('A[c] + B[c] => ')})
        g3 = vis.add_biomass_rxns(self.g, biorxn)
        bio_a = graph.Node({
            'id': 'test_bio_1', 'entry': [biorxn], 'shape': 'box',
            'style': 'filled', 'label': 'test_bio', 'type': 'bio_rxn',
            'fillcolor': '#b3fcb8', 'compartment': 'c'})
        bio_b = graph.Node({
            'id': 'test_bio_2', 'entry': [biorxn], 'shape': 'box',
            'style': 'filled', 'label': 'test_bio', 'type': 'bio_rxn',
            'fillcolor': '#b3fcb8', 'compartment': 'c'})
        edge_bio_a = graph.Edge(self.node_a, bio_a, {'dir': 'forward'})
        edge_bio_b = graph.Edge(self.node_b, bio_b, {'dir': 'forward'})
        final_nodes = [self.node_a, self.node_b, self.node_c, self.node_d,
                       self.node_rxn1, bio_a, bio_b]
        final_edges = [self.edge_A_rxn, self.edge_C_rxn,
                       edge_bio_a, edge_bio_b]
        self.assertTrue(all(i in final_nodes for i in g3.nodes))
        self.assertTrue(all(i in g3.nodes for i in final_nodes))
        self.assertTrue(all(i in final_edges for i in g3.edges))
        self.assertTrue(all(i in g3.edges for i in final_edges))


class TestAddExchange(unittest.TestCase):
    def setUp(self):
        native_model = NativeModel()
        native_model.reactions.add_entry(ReactionEntry({
            'id': 'rxn1', 'equation': parse_reaction(
                'fum[c] + h2o[c] <=> mal_L[c]')}))
        self.cpd_a = CompoundEntry({
            'id': 'A', 'name': 'compound A', 'formula': 'CO2', 'charge': 0})
        cpd_c = CompoundEntry({
            'id': 'C', 'name': 'compound C', 'formula': 'MNO', 'charge': 0})
        rxn1 = entry.DictReactionEntry({
            'id': 'rxn1', 'name': 'Reaction 1', 'equation': Reaction(
                Direction.Forward, {Compound('A', 'c'): -1,
                                    Compound('C', 'c'): 1})
        })
        rxn2 = entry.DictReactionEntry({
            'id': 'rxn2', 'name': 'Reaction 2', 'equation': Reaction(
                Direction.Forward, {Compound('A', 'e'): -1,
                                    Compound('A', 'c'): 1}),
        })
        rxn3 = entry.DictReactionEntry({
            'id': 'rxn3', 'name': 'Reaction 3', 'equation': Reaction(
                Direction.Forward, {Compound('C', 'c'): 1,
                                    Compound('C', 'e'): 1}),
        })

        node_a = graph.Node({
            'id': 'A[c]', 'entry': [self.cpd_a],
            'compartment': 'c', 'type': 'cpd',
            'style': 'filled', 'shape': 'ellipse', 'fillcolor': '#ffd8bf',
            'label': 'A[c]'})
        node_c = graph.Node({
            'id': 'C[c]', 'entry': [cpd_c], 'compartment': 'c', 'type': 'cpd',
            'style': 'filled', 'shape': 'ellipse', 'fillcolor': '#ffd8bf',
            'label': 'C[c]'})
        self.node_a_extracell = graph.Node({
            'id': 'A[e]', 'entry': [self.cpd_a],
            'compartment': 'e', 'type': 'cpd',
            'style': 'filled', 'shape': 'ellipse', 'fillcolor': '#ffd8bf',
            'label': 'A[e]'})
        self.node_c_extracell = graph.Node({
            'id': 'C[e]', 'entry': [cpd_c], 'compartment': 'e', 'type': 'cpd',
            'style': 'filled', 'shape': 'ellipse', 'fillcolor': '#ffd8bf',
            'label': 'C[e]'})
        node_ac = graph.Node({
            'id': 'rxn1_1', 'shape': 'box', 'style': 'filled', 'type': 'rxn',
            'entry': [rxn1], 'compartment': 'c',
            'fillcolor': '#c9fccd', 'label': 'rxn1'})
        self.node_aa = graph.Node({
            'id': 'rxn2_1', 'shape': 'box', 'style': 'filled', 'type': 'rxn',
            'entry': [rxn2], 'compartment': 'c', 'fillcolor': '#90f998',
            'label': 'rxn2'})
        node_cc = graph.Node({
            'id': 'rxn3_1', 'shape': 'box', 'style': 'filled', 'type': 'rxn',
            'entry': [rxn3], 'compartment': 'e', 'fillcolor': '#90f998',
            'label': 'rxn3'})
        edge_a_r1 = graph.Edge(node_a, node_ac, {'dir': 'both'})
        edge_r1_c = graph.Edge(node_ac, node_c, {'dir': 'both'})
        self.edge_a_r2 = graph.Edge(self.node_a_extracell,
                                    self.node_aa, {'dir': 'forward'})
        self.edge_r2_a = graph.Edge(self.node_aa, node_a, {'dir': 'forward'})
        edge_c_r3 = graph.Edge(self.node_c_extracell, node_cc, {'dir': 'back'})
        edge_r3_c = graph.Edge(node_cc, node_c, {'dir': 'back'})

        self.g = graph.Graph()
        self.edge_list = [edge_a_r1, edge_r1_c, self.edge_a_r2, self.edge_r2_a,
                          edge_c_r3, edge_r3_c]
        self.node_list = [node_a, node_c, self.node_a_extracell,
                          self.node_c_extracell, node_ac,
                          self.node_aa, node_cc]
        for node in self.node_list:
            self.g.add_node(node)
        for edge in self.edge_list:
            self.g.add_edge(edge)
        self.EX_A = Reaction(Direction.Both, {Compound('A', 'e'): -5})
        self.EX_C = Reaction(Direction.Both, {Compound('C', 'e'): 5})

    def test1_addExrRxn_cpdA(self):
        g1 = vis.add_exchange_rxns(self.g, 'test_EX_A', self.EX_A,
                                   {'test_EX_A': ('solid', 1)})
        node_ex = graph.Node({
            'id': 'test_EX_A', 'entry': [self.EX_A], 'shape': 'box',
            'style': 'filled', 'label': 'test_EX_A', 'type': 'Ex_rxn',
            'fillcolor': '#90f998', 'compartment': 'e'})
        edge_ex = graph.Edge(self.node_a_extracell, node_ex,
                             {'dir': 'both', 'style': 'solid',
                              'penwidth': 1})
        self.node_list.append(node_ex)
        self.edge_list.append(edge_ex)

        self.assertTrue(all(i in self.node_list for i in g1.nodes))
        self.assertTrue(all(i in g1.nodes for i in self.node_list))

        self.assertTrue(all(i in self.edge_list for i in g1.edges))
        self.assertTrue(all(i in g1.edges for i in self.edge_list))

    def test2_addExrRxn_right(self):
        g2 = vis.add_exchange_rxns(self.g, 'test_EX_C', self.EX_C,
                                   {'test_EX_C': ('solid', 1)})
        node_ex = graph.Node({
            'id': 'test_EX_C', 'entry': [self.EX_C], 'shape': 'box',
            'style': 'filled', 'label': 'test_EX_C', 'type': 'Ex_rxn',
            'fillcolor': '#90f998', 'compartment': 'e'})
        edge_ex = graph.Edge(node_ex, self.node_c_extracell,
                             {'dir': 'both', 'style': 'solid',
                              'penwidth': 1})
        self.node_list.append(node_ex)
        self.edge_list.append(edge_ex)

        self.assertTrue(all(i in self.node_list for i in g2.nodes))
        self.assertTrue(all(i in g2.nodes for i in self.node_list))

        self.assertTrue(all(i in self.edge_list for i in g2.edges))
        self.assertTrue(all(i in g2.edges for i in self.edge_list))

    def test3_addExrCpd(self):
        cpd_formula = {'A': Formula({Atom('O'): 2, Atom('C'): 1})}
        mm_a = Compound('A', 'e')
        new_node_a_extra = graph.Node({'id': 'A[e]', 'entry': [self.cpd_a],
                                       'compartment': 'e', 'type': 'cpd'})
        g_missA = graph.Graph()
        node_list_missA = [n for n in self.node_list if n not in [
            self.node_aa, self.node_a_extracell]]
        edge_list_missA = [e for e in self.edge_list if e not in [
            self.edge_a_r2, self.edge_r2_a]]
        for node in node_list_missA:
            g_missA.add_node(node)
        for edge in edge_list_missA:
            g_missA.add_edge(edge)
        g3 = vis.add_ex_cpd(g_missA, mm_a, self.cpd_a, cpd_formula, 'C')
        node_list_missA.append(new_node_a_extra)

        # for i in g3.nodes:
        #     print(i, i.props)
        # print('fen ge')
        # for i in node_list_missA:
        #     print(i, i.props)

        self.assertTrue(all(i in node_list_missA for i in g3.nodes))
        self.assertTrue(all(i in g3.nodes for i in node_list_missA))

        self.assertTrue(all(i in edge_list_missA for i in g3.edges))
        self.assertTrue(all(i in g3.edges for i in edge_list_missA))

    def test4_addExrCpd_noEle(self):
        cpd_formula = {'A': Formula({Atom('O'): 2, Atom('C'): 1})}
        mm_a = Compound('A', 'e')

        g_missA = graph.Graph()
        node_list_missA = [n for n in self.node_list if n not in [
            self.node_aa, self.node_a_extracell]]
        edge_list_missA = [e for e in self.edge_list if e not in [
            self.edge_a_r2, self.edge_r2_a]]
        for node in node_list_missA:
            g_missA.add_node(node)
        for edge in edge_list_missA:
            g_missA.add_edge(edge)
        g4 = vis.add_ex_cpd(g_missA, mm_a, self.cpd_a, cpd_formula, 'N')

        self.assertTrue(all(i in node_list_missA for i in g4.nodes))
        self.assertTrue(all(i in g4.nodes for i in node_list_missA))

        self.assertTrue(all(i in edge_list_missA for i in g4.edges))
        self.assertTrue(all(i in g4.edges for i in edge_list_missA))

    def test5_addExrCpd_noFormula(self):
        cpd_formula = {}
        mm_a = Compound('A', 'e')

        g_missA = graph.Graph()
        node_list_missA = [n for n in self.node_list if n not in [
            self.node_aa, self.node_a_extracell]]
        edge_list_missA = [e for e in self.edge_list if e not in [
            self.edge_a_r2, self.edge_r2_a]]
        for node in node_list_missA:
            g_missA.add_node(node)
        for edge in edge_list_missA:
            g_missA.add_edge(edge)
        g5 = vis.add_ex_cpd(g_missA, mm_a, self.cpd_a, cpd_formula, 'C')

        self.assertTrue(all(i in node_list_missA for i in g5.nodes))
        self.assertTrue(all(i in g5.nodes for i in node_list_missA))

        self.assertTrue(all(i in edge_list_missA for i in g5.edges))
        self.assertTrue(all(i in g5.edges for i in edge_list_missA))

    def test6_addExrCpd_EleNone(self):
        cpd_formula = {'A': Formula({Atom('O'): 2, Atom('C'): 1})}
        mm_a = Compound('A', 'e')
        new_node_a_extra = graph.Node({'id': 'A[e]', 'entry': [self.cpd_a],
                                       'compartment': 'e', 'type': 'cpd'})
        g_missA = graph.Graph()
        node_list_missA = [n for n in self.node_list if n not in [
            self.node_aa, self.node_a_extracell]]
        edge_list_missA = [e for e in self.edge_list if e not in [
            self.edge_a_r2, self.edge_r2_a]]
        for node in node_list_missA:
            g_missA.add_node(node)
        for edge in edge_list_missA:
            g_missA.add_edge(edge)
        g6 = vis.add_ex_cpd(g_missA, mm_a, self.cpd_a, cpd_formula, None)
        node_list_missA.append(new_node_a_extra)

        self.assertTrue(all(i in node_list_missA for i in g6.nodes))
        self.assertTrue(all(i in g6.nodes for i in node_list_missA))

        self.assertTrue(all(i in edge_list_missA for i in g6.edges))
        self.assertTrue(all(i in g6.edges for i in edge_list_missA))


class TestMakeCptTree(unittest.TestCase):
    def test1_if_e_inthemodel(self):
        boundaries = [('c', 'e'), ('c', 'p'), ('e', 'p')]
        extra = 'e'
        c1, e1 = vis.make_cpt_tree(boundaries, extra)
        c1_res = defaultdict(set)
        c1_res['c'].add('e')
        c1_res['c'].add('p')
        c1_res['p'].add('e')
        c1_res['p'].add('c')
        c1_res['e'].add('c')
        c1_res['e'].add('p')
        e1_res = 'e'
        self.assertEqual(c1, c1_res)
        self.assertEqual(e1, e1_res)

    def test2_if_e_not_inthemodel(self):
        boundaries = [('c', 'mi')]
        extra = 'e'
        c2, e2 = vis.make_cpt_tree(boundaries, extra)
        c2_res = defaultdict(set)
        c2_res['c'].add('mi')
        c2_res['mi'].add('c')
        e2_res = 'mi'
        self.assertEqual(c2, c2_res)
        self.assertEqual(e2, e2_res)

    def test3_if_e_not_inthemodel_cpmi(self):
        boundaries = [('c', 'p'), ('c', 'mi')]
        extra = 'e'
        c3, e3 = vis.make_cpt_tree(boundaries, extra)
        c3_res = defaultdict(set)
        c3_res['c'].add('p')
        c3_res['c'].add('mi')
        c3_res['mi'].add('c')
        c3_res['p'].add('c')
        e3_res = 'p'
        self.assertEqual(c3, c3_res)
        self.assertEqual(e3, e3_res)


class TestGetCptBoundaries(unittest.TestCase):
    def setUp(self):
        native_model = NativeModel()
        native_model.reactions.add_entry(ReactionEntry({
            'id': 'rxn1', 'equation': parse_reaction('A[c] => A[e]')}))
        native_model.compounds.add_entry(CompoundEntry({
            'id': 'A[c]', 'formula': parse_compound('formula_A', 'c')}))
        native_model.compounds.add_entry(CompoundEntry({
            'id': 'A[e]', 'formula': parse_compound('formula_A', 'e')}))
        native_model.reactions.add_entry(ReactionEntry({
            'id': 'rxn2', 'equation': parse_reaction('B[e] <=> B[p]')}))
        native_model.compounds.add_entry(CompoundEntry({
            'id': 'B[e]', 'formula': parse_compound('formula_B', 'e')}))
        native_model.compounds.add_entry(CompoundEntry({
            'id': 'B[p]', 'formula': parse_compound('formula_B', 'p')}))
        self.native = native_model

    def test1_default_setting(self):
        e1_bound, e1_extra = vis.get_cpt_boundaries(self.native)
        e1_bound_res = set()
        e1_bound_res.add(('c', 'e'))
        e1_bound_res.add(('e', 'p'))
        e1_extra_res = 'e'
        self.assertTrue(all(i in e1_bound for i in e1_bound_res))
        self.assertEqual(e1_extra, e1_extra_res)

    def test2_with_extra_defined(self):
        self.native._properties['extracellular'] = 'p'
        e2_bound, e2_extra = vis.get_cpt_boundaries(self.native)
        e2_bound_res = set()
        e2_bound_res.add(('c', 'e'))
        e2_bound_res.add(('e', 'p'))
        e2_extra_res = 'p'
        self.assertTrue(all(i in e2_bound for i in e2_bound_res))
        self.assertEqual(e2_extra, e2_extra_res)

    def test3_e_notinmodel(self):
        native_model = NativeModel()
        native_model.reactions.add_entry(ReactionEntry({
            'id': 'rxn1', 'equation': parse_reaction('A[c] => A[mi]')}))
        e3_bound, e3_extra = vis.get_cpt_boundaries(native_model)
        e3_bound_res = set()
        e3_bound_res.add(('c', 'mi'))
        e3_extra_res = 'e'
        self.assertEqual(e3_bound, e3_bound_res)
        self.assertEqual(e3_extra, e3_extra_res)

    def test4_cpt_boundaries_defined_in_model(self):
        self.native.compartment_boundaries.add(('c', 'e'))
        self.native.compartment_boundaries.add(('e', 'p'))
        self.native.compartment_boundaries.add(('c', 'mi'))
        e4_bound, e4_extra = vis.get_cpt_boundaries(self.native)
        e4_bound_res = set()
        e4_bound_res.add(('c', 'e'))
        e4_bound_res.add(('e', 'p'))
        e4_bound_res.add(('c', 'mi'))
        e4_extra_res = 'e'
        self.assertTrue(all(i in e4_bound for i in e4_bound_res))
        self.assertEqual(e4_extra, e4_extra_res)


class TestFlux(unittest.TestCase):
    def setUp(self):
        native_model = NativeModel()
        native_model.reactions.add_entry(ReactionEntry({
            'id': 'rxn1', 'equation': parse_reaction(
                'fum[c] + h2o[c] <=> mal_L[c]')}))
        native_model.compounds.add_entry(CompoundEntry({
            'id': 'fum', 'formula': 'C4H2O4'}))
        native_model.compounds.add_entry(CompoundEntry({
            'id': 'h2o', 'formula': 'H2O'}))
        native_model.compounds.add_entry(CompoundEntry(
            {'id': 'mal_L', 'formula': 'C4H4O5'}))
        native_model.reactions.add_entry(ReactionEntry({
            'id': 'rxn2', 'equation': parse_reaction(
                'q8[c] + succ[c] => fum[c] + q8h2[c]')}))
        native_model.compounds.add_entry(CompoundEntry(
            {'id': 'q8', 'formula': 'C49H74O4'}))
        native_model.compounds.add_entry(CompoundEntry({
            'id': 'q8h2', 'formula': 'C49H76O4'}))
        native_model.compounds.add_entry(CompoundEntry({
            'id': 'succ', 'formula': 'C4H4O4'}))
        native_model.reactions.add_entry(ReactionEntry({
            'id': 'rxn3', 'equation': parse_reaction(
                'h2o[c] <=> h2o[e]')}))
        native_model.compounds.add_entry(CompoundEntry({
            'id': 'h2o', 'formula': 'H2O'}))
        self.native = native_model
        self.mm = native_model.create_metabolic_model()

    def test1_neg_flux_fba(self):
        reaction_dict = {'rxn1': (-20, 1)}
        full_pairs_dict, style_flux_dict = \
            graph.make_network_dict(self.native, self.mm,
                                    reaction_dict=reaction_dict,
                                    analysis='fba')
        self.assertEqual(Direction.Reverse,
                         full_pairs_dict[self.native.reactions['rxn1']][1])
        self.assertEqual('solid', style_flux_dict['rxn1'][0])
        self.assertEqual(5, style_flux_dict['rxn1'][1])

    def test2_pos_flux_fba(self):
        reaction_dict = {'rxn1': (10, 1)}
        full_pairs_dict, style_flux_dict = \
            graph.make_network_dict(self.native, self.mm,
                                    reaction_dict=reaction_dict,
                                    analysis='fba')
        self.assertEqual(Direction.Forward,
                         full_pairs_dict[self.native.reactions['rxn1']][1])
        self.assertEqual('solid', style_flux_dict['rxn1'][0])
        self.assertEqual(5, style_flux_dict['rxn1'][1])

    def test3_0_flux_fba(self):
        reaction_dict = {'rxn1': (0, 1)}
        full_pairs_dict, style_flux_dict = \
            graph.make_network_dict(self.native, self.mm,
                                    reaction_dict=reaction_dict,
                                    analysis='fba')
        self.assertEqual(Direction.Both,
                         full_pairs_dict[self.native.reactions['rxn1']][1])
        self.assertEqual('dotted', style_flux_dict['rxn1'][0])
        self.assertEqual(1, style_flux_dict['rxn1'][1])

    def test4_neg_flux_fva(self):
        reaction_dict = {'rxn1': (-10, 1)}
        full_pairs_dict, style_flux_dict = \
            graph.make_network_dict(self.native, self.mm,
                                    reaction_dict=reaction_dict,
                                    analysis='fva')
        self.assertEqual(Direction.Both,
                         full_pairs_dict[self.native.reactions['rxn1']][1])
        self.assertEqual('dotted', style_flux_dict['rxn2'][0])
        self.assertEqual(5, style_flux_dict['rxn1'][1])

    def test5_pos_flux_neg_fva(self):
        reaction_dict = {'rxn1': (-10, -10)}
        full_pairs_dict, style_flux_dict = \
            graph.make_network_dict(self.native, self.mm,
                                    reaction_dict=reaction_dict,
                                    analysis='fva')
        self.assertEqual(Direction.Reverse,
                         full_pairs_dict[self.native.reactions['rxn1']][1])
        self.assertEqual('solid', style_flux_dict['rxn1'][0])
        self.assertEqual(5, style_flux_dict['rxn1'][1])

    def test6_pos_flux_pos_fva(self):
        reaction_dict = {'rxn1': (10, 10)}
        full_pairs_dict, style_flux_dict = \
            graph.make_network_dict(self.native, self.mm,
                                    reaction_dict=reaction_dict,
                                    analysis='fva')
        self.assertEqual(Direction.Forward,
                         full_pairs_dict[self.native.reactions['rxn1']][1])
        self.assertEqual('solid', style_flux_dict['rxn1'][0])
        self.assertEqual(5, style_flux_dict['rxn1'][1])

    def test7_lower_0_flux_fva(self):
        reaction_dict = {'rxn1': (0, 10)}
        full_pairs_dict, style_flux_dict = \
            graph.make_network_dict(self.native, self.mm,
                                    reaction_dict=reaction_dict,
                                    analysis='fva')
        self.assertEqual(Direction.Forward,
                         full_pairs_dict[self.native.reactions['rxn1']][1])
        self.assertEqual('solid', style_flux_dict['rxn1'][0])
        self.assertEqual(1, style_flux_dict['rxn1'][1])

    def test8_upper_0_flux_fva(self):
        reaction_dict = {'rxn1': (-10, 0)}
        full_pairs_dict, style_flux_dict = \
            graph.make_network_dict(self.native, self.mm,
                                    reaction_dict=reaction_dict,
                                    analysis='fva')
        self.assertEqual(Direction.Reverse,
                         full_pairs_dict[self.native.reactions['rxn1']][1])
        self.assertEqual('solid', style_flux_dict['rxn1'][0])
        self.assertEqual(1, style_flux_dict['rxn1'][1])

    def test9_both_0_flux_fva(self):
        reaction_dict = {'rxn1': (0, 0)}
        full_pairs_dict, style_flux_dict = \
            graph.make_network_dict(self.native, self.mm,
                                    reaction_dict=reaction_dict,
                                    analysis='fva')
        self.assertEqual(Direction.Both,
                         full_pairs_dict[self.native.reactions['rxn1']][1])
        self.assertEqual('dotted', style_flux_dict['rxn1'][0])
        self.assertEqual(1, style_flux_dict['rxn1'][1])

    def test10_mult_rxn_fba(self):
        reaction_dict = {'rxn1': (2, 1), 'rxn2': (-4, 1), 'rxn3': (6, 1)}
        full_pairs_dict, style_flux_dict = \
            graph.make_network_dict(self.native, self.mm,
                                    reaction_dict=reaction_dict,
                                    analysis='fba')
        self.assertEqual(Direction.Forward,
                         full_pairs_dict[self.native.reactions['rxn1']][1])
        self.assertEqual(Direction.Reverse,
                         full_pairs_dict[self.native.reactions['rxn2']][1])
        self.assertEqual(Direction.Forward,
                         full_pairs_dict[self.native.reactions['rxn3']][1])
        self.assertEqual('solid', style_flux_dict['rxn1'][0])
        self.assertEqual('solid', style_flux_dict['rxn2'][0])
        self.assertEqual('solid', style_flux_dict['rxn3'][0])
        self.assertEqual(2.5, style_flux_dict['rxn1'][1])
        self.assertEqual(5, style_flux_dict['rxn2'][1])
        self.assertEqual(7.5, style_flux_dict['rxn3'][1])

    def test11_mult_rxn_fva(self):
        reaction_dict = {'rxn1': (0, 2), 'rxn2': (-4, 1), 'rxn3': (-1, -1)}
        full_pairs_dict, style_flux_dict = \
            graph.make_network_dict(self.native, self.mm,
                                    reaction_dict=reaction_dict,
                                    analysis='fva')
        self.assertEqual(Direction.Forward,
                         full_pairs_dict[self.native.reactions['rxn1']][1])
        self.assertEqual(Direction.Both,
                         full_pairs_dict[self.native.reactions['rxn2']][1])
        self.assertEqual(Direction.Reverse,
                         full_pairs_dict[self.native.reactions['rxn3']][1])
        self.assertEqual('solid', style_flux_dict['rxn1'][0])
        self.assertEqual('solid', style_flux_dict['rxn2'][0])
        self.assertEqual('solid', style_flux_dict['rxn3'][0])
        self.assertEqual(1, style_flux_dict['rxn1'][1])
        self.assertEqual(8, style_flux_dict['rxn2'][1])
        self.assertEqual(2, style_flux_dict['rxn3'][1])


if __name__ == '__main__':
    unittest.main()
